# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import argparse
import logging
import signal

from botocore.exceptions import (
    ClientError,
    NoCredentialsError,
    NoRegionError,
)
from botocore.exceptions import (
    ParamValidationError as BotocoreParamValidationError,
)

from awscli.argparser import USAGE, ArgParseException
from awscli.argprocess import ParamError, ParamSyntaxError
from awscli.arguments import UnknownArgumentError
from awscli.autoprompt.factory import PrompterKeyboardInterrupt
from awscli.constants import (
    CLIENT_ERROR_RC,
    CONFIGURATION_ERROR_RC,
    GENERAL_ERROR_RC,
    PARAM_VALIDATION_ERROR_RC,
)
from awscli.customizations.exceptions import (
    ConfigurationError,
    ParamValidationError,
)
from awscli.errorformat import write_error
from awscli.formatter import get_formatter
from awscli.utils import PagerInitializationException

LOG = logging.getLogger(__name__)

VALID_ERROR_FORMATS = ['legacy', 'json', 'yaml', 'text', 'table', 'enhanced']
# Maximum number of items to display inline for collections
MAX_INLINE_ITEMS = 5


class EnhancedErrorFormatter:
    def format_error(self, error_info, stream):
        additional_fields = self._get_additional_fields(error_info)

        if not additional_fields:
            return

        stream.write('\nAdditional error details:\n')
        has_complex_value = False
        for key, value in additional_fields.items():
            if self._is_simple_value(value):
                stream.write(f'{key}: {value}\n')
            elif self._is_small_collection(value):
                stream.write(f'{key}: {self._format_inline(value)}\n')
            else:
                stream.write(f'{key}: <complex value>\n')
                has_complex_value = True

        if has_complex_value:
            stream.write(
                'Use "--cli-error-format json" or another error format '
                'to see the full details.\n'
            )

    def _is_simple_value(self, value):
        return isinstance(value, (str, int, float, bool, type(None)))

    def _is_small_collection(self, value):
        if isinstance(value, list):
            return len(value) < MAX_INLINE_ITEMS and all(
                self._is_simple_value(item) for item in value
            )
        elif isinstance(value, dict):
            return len(value) < MAX_INLINE_ITEMS and all(
                self._is_simple_value(v) for v in value.values()
            )
        return False

    def _format_inline(self, value):
        if isinstance(value, list):
            return f"[{', '.join(str(item) for item in value)}]"
        elif isinstance(value, dict):
            items = ', '.join(f'{k}: {v}' for k, v in value.items())
            return f'{{{items}}}'
        return str(value)

    def _get_additional_fields(self, error_info):
        standard_keys = {'code', 'message'}
        return {
            k: v
            for k, v in error_info.items()
            if k.lower() not in standard_keys
        }


def construct_entry_point_handlers_chain():
    handlers = [
        ParamValidationErrorsHandler(),
        PrompterInterruptExceptionHandler(),
        InterruptExceptionHandler(),
        GeneralExceptionHandler(),
    ]
    return ChainedExceptionHandler(exception_handlers=handlers)


def construct_cli_error_handlers_chain(session=None):
    # UnknownArgumentErrorHandler and InterruptExceptionHandler are
    # intentionally excluded from structured formatting
    handlers = [
        ParamValidationErrorsHandler(session),
        UnknownArgumentErrorHandler(),
        ConfigurationErrorHandler(session),
        NoRegionErrorHandler(session),
        NoCredentialsErrorHandler(session),
        PagerErrorHandler(session),
        InterruptExceptionHandler(),
        ClientErrorHandler(session),
        GeneralExceptionHandler(session),
    ]
    return ChainedExceptionHandler(exception_handlers=handlers)


class BaseExceptionHandler:
    def handle_exception(self, exception, stdout, stderr, **kwargs):
        raise NotImplementedError('handle_exception')


class FilteredExceptionHandler(BaseExceptionHandler):
    EXCEPTIONS_TO_HANDLE = ()
    RC = None

    def __init__(self, session=None):
        self._session = session

    def handle_exception(self, exception, stdout, stderr, **kwargs):
        if isinstance(exception, self.EXCEPTIONS_TO_HANDLE):
            return_val = self._do_handle_exception(
                exception, stdout, stderr, **kwargs
            )
            if return_val is not None:
                return return_val

    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        parsed_globals = kwargs.get('parsed_globals')
        error_info = self._extract_error_info(exception)

        if error_info:
            formatted_message = self._get_formatted_message(
                error_info, exception
            )
            displayed_structured = self._display_structured_error(
                error_info, formatted_message, stderr, parsed_globals
            )
            if displayed_structured:
                return self.RC

        message = (error_info or {}).get('Message', str(exception))
        write_error(stderr, message)
        return self.RC

    def _extract_error_info(self, exception):
        """Extract error information for structured formatting.

        Returns None by default. Subclasses should override to provide
        error information as a dict with 'Code' and 'Message' keys.
        """
        return None

    def _get_formatted_message(self, error_info, exception):
        code = error_info.get('Code', 'Unknown')
        message = error_info.get('Message', str(exception))
        return f"An error occurred ({code}): {message}"

    def _resolve_error_format(self, parsed_globals):
        if parsed_globals:
            error_format = getattr(parsed_globals, 'cli_error_format', None)
            if error_format:
                return error_format.lower()

        if self._session:
            try:
                error_format = self._session.get_config_variable(
                    'cli_error_format'
                )
                if error_format:
                    return error_format.lower()
            except (KeyError, AttributeError) as e:
                LOG.debug('Failed to get cli_error_format from config: %s', e)

        return 'enhanced'

    def _display_structured_error(
        self, error_info, formatted_message, stderr, parsed_globals=None
    ):
        try:
            error_format = self._resolve_error_format(parsed_globals)

            if error_format == 'legacy':
                return False

            if error_format not in VALID_ERROR_FORMATS:
                LOG.warning(
                    f"Invalid cli_error_format: '{error_format}'. "
                    f"Using 'enhanced' format."
                )
                error_format = 'enhanced'

            if error_format == 'enhanced':
                write_error(stderr, formatted_message)
                EnhancedErrorFormatter().format_error(error_info, stderr)
                return True

            color = getattr(parsed_globals, 'color', 'auto')
            formatter_args = argparse.Namespace(query=None, color=color)
            formatter = get_formatter(error_format, formatter_args)
            formatter('error', error_info, stderr)
            return True

        except Exception as e:
            LOG.debug(
                'Failed to display structured error: %s', e, exc_info=True
            )
            return False


class ParamValidationErrorsHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = (
        ParamError,
        ParamSyntaxError,
        ArgParseException,
        ParamValidationError,
        BotocoreParamValidationError,
    )
    RC = PARAM_VALIDATION_ERROR_RC

    def _extract_error_info(self, exception):
        return {'Code': 'ParamValidation', 'Message': str(exception)}


class SilenceParamValidationMsgErrorHandler(ParamValidationErrorsHandler):
    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        return self.RC


class ClientErrorHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = ClientError
    RC = CLIENT_ERROR_RC

    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        parsed_globals = kwargs.get('parsed_globals')
        error_info = self._extract_error_info(exception)

        if error_info:
            formatted_message = self._get_formatted_message(
                error_info, exception
            )
            displayed_structured = self._display_structured_error(
                error_info, formatted_message, stderr, parsed_globals
            )
            if displayed_structured:
                return self.RC

        write_error(stderr, str(exception))
        return self.RC

    def _get_formatted_message(self, error_info, exception):
        return str(exception)

    def _extract_error_info(self, exception):
        error_response = self._extract_error_response(exception)
        if error_response and 'Error' in error_response:
            return error_response['Error']
        return None

    @staticmethod
    def _extract_error_response(exception):
        if not isinstance(exception, ClientError):
            return None

        if hasattr(exception, 'response') and 'Error' in exception.response:
            error_dict = dict(exception.response['Error'])

            modeled_fields = exception.modeled_fields
            if modeled_fields is not None:
                modeled_lower = {f.lower() for f in modeled_fields}
            else:
                modeled_lower = {'code', 'message'}

            # Only include fields present in the error shape model.
            excluded_keys = {'Error', 'ResponseMetadata', 'Code', 'Message'}
            for key, value in exception.response.items():
                if key not in excluded_keys and key not in error_dict:
                    if key.lower() in modeled_lower:
                        error_dict[key] = value

            # Unmodeled fields can contain data that shouldn't be displayed.
            for key in list(error_dict.keys()):
                if key.lower() not in modeled_lower:
                    del error_dict[key]

            return {'Error': error_dict}

        return None


class ConfigurationErrorHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = ConfigurationError
    RC = CONFIGURATION_ERROR_RC

    def _extract_error_info(self, exception):
        return {'Code': 'Configuration', 'Message': str(exception)}


class NoRegionErrorHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = NoRegionError
    RC = CONFIGURATION_ERROR_RC

    def _extract_error_info(self, exception):
        message = (
            f'{exception} You can also configure your region by running '
            f'"aws configure".'
        )
        return {'Code': 'NoRegion', 'Message': message}


class NoCredentialsErrorHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = NoCredentialsError
    RC = CONFIGURATION_ERROR_RC

    def _extract_error_info(self, exception):
        message = (
            f'{exception}. You can configure credentials '
            f'by running "aws login".'
        )
        return {'Code': 'NoCredentials', 'Message': message}


class PagerErrorHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = PagerInitializationException
    RC = CONFIGURATION_ERROR_RC

    def _extract_error_info(self, exception):
        message = (
            f'Unable to redirect output to pager. Received the '
            f'following error when opening pager:\n{exception}\n\n'
            f'Learn more about configuring the output pager by running '
            f'"aws help config-vars".'
        )
        return {'Code': 'Pager', 'Message': message}


class UnknownArgumentErrorHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = UnknownArgumentError
    RC = PARAM_VALIDATION_ERROR_RC

    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        stderr.write("\n")
        stderr.write(f'usage: {USAGE}\n')
        write_error(stderr, str(exception))
        return self.RC


class InterruptExceptionHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = KeyboardInterrupt
    RC = 128 + signal.SIGINT

    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        stdout.write("\n")
        return self.RC


class PrompterInterruptExceptionHandler(InterruptExceptionHandler):
    EXCEPTIONS_TO_HANDLE = PrompterKeyboardInterrupt

    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        stderr.write(f'{exception}')
        stderr.write("\n")
        return self.RC


class GeneralExceptionHandler(FilteredExceptionHandler):
    EXCEPTIONS_TO_HANDLE = Exception
    RC = GENERAL_ERROR_RC

    def _do_handle_exception(self, exception, stdout, stderr, **kwargs):
        # Generic exceptions don't have meaningful structure,
        # so always use plain text formatting
        write_error(stderr, str(exception))
        return self.RC


class ChainedExceptionHandler(BaseExceptionHandler):
    def __init__(self, exception_handlers):
        self._exception_handlers = exception_handlers

    def inject_handler(self, position, handler):
        self._exception_handlers.insert(position, handler)

    def handle_exception(self, exception, stdout, stderr, **kwargs):
        for handler in self._exception_handlers:
            return_value = handler.handle_exception(
                exception, stdout, stderr, **kwargs
            )
            if return_value is not None:
                return return_value



import asyncio
import re

from prompt_toolkit.application import Application
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import HSplit, Layout, ScrollablePane, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.widgets import Frame


def _get_visual_line_length(line):
    """Get the visual length of a line, excluding ANSI escape codes."""
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return len(ansi_escape.sub('', line))


class Display:
    def __init__(self):
        self.control = FormattedTextControl(text="")
        self.window = ScrollablePane(
            Window(
                content=self.control, wrap_lines=True, dont_extend_width=False
            ),
            show_scrollbar=True,
        )
        self.status_control = FormattedTextControl(
            text="up/down to scroll, q to quit"
        )
        self.content_lines = 0
        self.raw_text = ""
        kb = KeyBindings()

        @kb.add('q')
        @kb.add('c-c')
        def quit_app(event):
            event.app.exit()

        @kb.add('up')
        def scroll_up(event):
            if self.window.vertical_scroll > 0:
                self.window.vertical_scroll -= 1

        @kb.add('down')
        def scroll_down(event):
            # Frame top, frame bottom, status bar
            window_height = event.app.output.get_size().rows - 3
            # Account for frame borders and scrollbar
            window_width = event.app.output.get_size().columns - 4

            total_display_lines = self._calculate_display_lines(window_width)
            max_scroll = max(0, total_display_lines - window_height)
            if self.window.vertical_scroll < max_scroll:
                self.window.vertical_scroll += 1

        self.app = Application(
            layout=Layout(
                HSplit(
                    [
                        Frame(self.window),
                        Window(content=self.status_control, height=1),
                    ]
                )
            ),
            key_bindings=kb,
            full_screen=True,
        )

    def display(self, text, status_text=""):
        """Update display with ANSI colored text."""
        self.raw_text = text
        self.control.text = ANSI(text)
        self.content_lines = len(text.split('\n'))

        self._validate_scroll_position()

        if status_text:
            self.status_control.text = status_text
        self.app.invalidate()

    def _calculate_display_lines(self, window_width):
        """Calculate total display lines accounting for line wrapping."""
        total_display_lines = 0
        for line in self.raw_text.split('\n'):
            visual_length = _get_visual_line_length(line)
            if visual_length == 0:
                total_display_lines += 1
            else:
                # Calculate how many display lines this text line will occupy
                # when wrapped: ceil(visual_length / window_width)
                # Using integer math: (visual_length + window_width - 1) // window_width
                total_display_lines += max(
                    1, (visual_length + window_width - 1) // window_width
                )
        return total_display_lines

    def _validate_scroll_position(self):
        """Ensure scroll position is valid for current content."""
        if not getattr(self.app, 'output', None):
            return

        try:
            window_height = self.app.output.get_size().rows - 3
            window_width = self.app.output.get_size().columns - 4

            total_display_lines = self._calculate_display_lines(window_width)
            max_scroll = max(0, total_display_lines - window_height)
            if self.window.vertical_scroll > max_scroll:
                self.window.vertical_scroll = max_scroll
        except (AttributeError, OSError):
            # If we can't determine terminal size, leave scroll position unchanged
            pass

    async def run(self):
        """Run the display app."""
        await self.app.run_async()



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from awscli.customizations.s3.syncstrategy.base import SizeAndLastModifiedSync

LOG = logging.getLogger(__name__)


EXACT_TIMESTAMPS = {
    'name': 'exact-timestamps',
    'action': 'store_true',
    'help_text': (
        'When syncing from S3 to local, same-sized '
        'items will be ignored only when the timestamps '
        'match exactly. The default behavior is to ignore '
        'same-sized items unless the local version is newer '
        'than the S3 version.'
    ),
}


class ExactTimestampsSync(SizeAndLastModifiedSync):
    ARGUMENT = EXACT_TIMESTAMPS

    def compare_time(self, src_file, dest_file):
        src_time = src_file.last_update
        dest_time = dest_file.last_update
        delta = dest_time - src_time
        cmd = src_file.operation_name
        if cmd == 'download':
            return self.total_seconds(delta) == 0
        else:
            return super(ExactTimestampsSync, self).compare_time(
                src_file, dest_file
            )



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging

from awscli.customizations.commands import BasicCommand
from awscli.customizations.emr import config, configutils, emrutils, exceptions

LOG = logging.getLogger(__name__)


class Command(BasicCommand):
    region = None

    UNSUPPORTED_COMMANDS_FOR_RELEASE_BASED_CLUSTERS = set(
        [
            'install-applications',
            'restore-from-hbase-backup',
            'schedule-hbase-backup',
            'create-hbase-backup',
            'disable-hbase-backups',
        ]
    )

    def supports_arg(self, name):
        return any(x['name'] == name for x in self.ARG_TABLE)

    def _run_main(self, parsed_args, parsed_globals):
        self._apply_configs(
            parsed_args, configutils.get_configs(self._session)
        )
        self.region = emrutils.get_region(self._session, parsed_globals)
        self._validate_unsupported_commands_for_release_based_clusters(
            parsed_args, parsed_globals
        )
        return self._run_main_command(parsed_args, parsed_globals)

    def _apply_configs(self, parsed_args, parsed_configs):
        applicable_configurations = self._get_applicable_configurations(
            parsed_args, parsed_configs
        )

        configs_added = {}
        for configuration in applicable_configurations:
            configuration.add(
                self, parsed_args, parsed_configs[configuration.name]
            )
            configs_added[configuration.name] = parsed_configs[
                configuration.name
            ]

        if configs_added:
            LOG.debug("Updated arguments with configs: %s" % configs_added)
        else:
            LOG.debug("No configs applied")
        LOG.debug("Running command with args: %s" % parsed_args)

    def _get_applicable_configurations(self, parsed_args, parsed_configs):
        # We need to find the applicable configurations by applying
        # following filters:
        # 1. Configurations that are applicable to this command
        # 3. Configurations that are present in parsed_configs
        # 2. Configurations that are not present in parsed_args

        configurations = config.get_applicable_configurations(self)

        configurations = [
            x
            for x in configurations
            if x.name in parsed_configs and not x.is_present(parsed_args)
        ]

        configurations = self._filter_configurations_in_special_cases(
            configurations, parsed_args, parsed_configs
        )

        return configurations

    def _filter_configurations_in_special_cases(
        self, configurations, parsed_args, parsed_configs
    ):
        # Subclasses can override this method to filter the applicable
        # configurations further based upon some custom logic
        # Default behavior is to return the configurations list as is
        return configurations

    def _run_main_command(self, parsed_args, parsed_globals):
        # Subclasses should implement this method.
        # parsed_globals are the parsed global args (things like region,
        # profile, output, etc.)
        # parsed_args are any arguments you've defined in your ARG_TABLE
        # that are parsed.
        # parsed_args are updated to include any emr specific configuration
        # from the config file if the corresponding argument is not
        # explicitly specified on the CLI
        raise NotImplementedError("_run_main_command")

    def _validate_unsupported_commands_for_release_based_clusters(
        self, parsed_args, parsed_globals
    ):
        command = self.NAME

        if (
            command in self.UNSUPPORTED_COMMANDS_FOR_RELEASE_BASED_CLUSTERS
            and hasattr(parsed_args, 'cluster_id')
        ):
            release_label = emrutils.get_release_label(
                parsed_args.cluster_id,
                self._session,
                self.region,
                parsed_globals.endpoint_url,
                parsed_globals.verify_ssl,
            )
            if release_label:
                raise exceptions.UnsupportedCommandWithReleaseError(
                    command=command, release_label=release_label
                )


def override_args_required_option(argument_table, args, session, **kwargs):
    # This function overrides the 'required' property of an argument
    # if a value corresponding to that argument is present in the config
    # file
    # We don't want to override when user is viewing the help so that we
    # can show the required options correctly in the help
    need_to_override = False if len(args) == 1 and args[0] == 'help' else True

    if need_to_override:
        parsed_configs = configutils.get_configs(session)
        for arg_name in argument_table.keys():
            if arg_name.replace('-', '_') in parsed_configs:
                argument_table[arg_name].required = False



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.argprocess import unpack_cli_arg
from awscli.arguments import CustomArgument, create_argument_model_from_schema
from awscli.customizations.exceptions import ParamValidationError

S3_LOCATION_ARG_DESCRIPTION = {
    'name': 's3-location',
    'required': False,
    'help_text': (
        'Information about the location of the application revision in Amazon '
        'S3. You must specify the bucket, the key, and bundleType. '
        'Optionally, you can also specify an eTag and version.'
    ),
}

S3_LOCATION_SCHEMA = {
    "type": "object",
    "properties": {
        "bucket": {
            "type": "string",
            "description": "The Amazon S3 bucket name.",
            "required": True,
        },
        "key": {
            "type": "string",
            "description": "The Amazon S3 object key name.",
            "required": True,
        },
        "bundleType": {
            "type": "string",
            "description": "The format of the bundle stored in Amazon S3.",
            "enum": ["tar", "tgz", "zip"],
            "required": True,
        },
        "eTag": {
            "type": "string",
            "description": "The Amazon S3 object eTag.",
            "required": False,
        },
        "version": {
            "type": "string",
            "description": "The Amazon S3 object version.",
            "required": False,
        },
    },
}

GITHUB_LOCATION_ARG_DESCRIPTION = {
    'name': 'github-location',
    'required': False,
    'help_text': (
        'Information about the location of the application revision in '
        'GitHub. You must specify the repository and commit ID that '
        'references the application revision. For the repository, use the '
        'format GitHub-account/repository-name or GitHub-org/repository-name. '
        'For the commit ID, use the SHA1 Git commit reference.'
    ),
}

GITHUB_LOCATION_SCHEMA = {
    "type": "object",
    "properties": {
        "repository": {
            "type": "string",
            "description": (
                "The GitHub account or organization and repository. Specify "
                "as GitHub-account/repository or GitHub-org/repository."
            ),
            "required": True,
        },
        "commitId": {
            "type": "string",
            "description": "The SHA1 Git commit reference.",
            "required": True,
        },
    },
}


def modify_revision_arguments(argument_table, session, **kwargs):
    s3_model = create_argument_model_from_schema(S3_LOCATION_SCHEMA)
    argument_table[S3_LOCATION_ARG_DESCRIPTION['name']] = S3LocationArgument(
        argument_model=s3_model, session=session, **S3_LOCATION_ARG_DESCRIPTION
    )
    github_model = create_argument_model_from_schema(GITHUB_LOCATION_SCHEMA)
    argument_table[GITHUB_LOCATION_ARG_DESCRIPTION['name']] = (
        GitHubLocationArgument(
            argument_model=github_model,
            session=session,
            **GITHUB_LOCATION_ARG_DESCRIPTION,
        )
    )
    argument_table['revision'].required = False


class LocationArgument(CustomArgument):
    def __init__(self, session, *args, **kwargs):
        super(LocationArgument, self).__init__(*args, **kwargs)
        self._session = session

    def add_to_params(self, parameters, value):
        if value is None:
            return
        parsed = self._session.emit_first_non_none_response(
            'process-cli-arg.codedeploy.%s' % self.name,
            param=self.argument_model,
            cli_argument=self,
            value=value,
            operation=None,
        )
        if parsed is None:
            parsed = unpack_cli_arg(self, value)
        parameters['revision'] = self.build_revision_location(parsed)

    def build_revision_location(self, value_dict):
        """
        Repack the input structure into a revisionLocation.
        """
        raise NotImplementedError("build_revision_location")


class S3LocationArgument(LocationArgument):
    def build_revision_location(self, value_dict):
        required = ['bucket', 'key', 'bundleType']
        valid = lambda k: value_dict.get(k, False)
        if not all(map(valid, required)):
            raise ParamValidationError(
                '--s3-location must specify bucket, key and bundleType.'
            )
        revision = {
            "revisionType": "S3",
            "s3Location": {
                "bucket": value_dict['bucket'],
                "key": value_dict['key'],
                "bundleType": value_dict['bundleType'],
            },
        }
        if 'eTag' in value_dict:
            revision['s3Location']['eTag'] = value_dict['eTag']
        if 'version' in value_dict:
            revision['s3Location']['version'] = value_dict['version']
        return revision


class GitHubLocationArgument(LocationArgument):
    def build_revision_location(self, value_dict):
        required = ['repository', 'commitId']
        valid = lambda k: value_dict.get(k, False)
        if not all(map(valid, required)):
            raise ParamValidationError(
                '--github-location must specify repository and commitId.'
            )
        return {
            "revisionType": "GitHub",
            "gitHubLocation": {
                "repository": value_dict['repository'],
                "commitId": value_dict['commitId'],
            },
        }



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Dimension,
    Float,
    FloatContainer,
    ScrollOffsets,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.layout.margins import ScrollbarMargin
from prompt_toolkit.layout.menus import CompletionsMenu

from awscli.customizations.wizard.ui.selectmenu import (
    CollapsableSelectionMenuControl,
)
from awscli.customizations.wizard.ui.utils import FullyExtendedWidthWindow


class WizardPrompt:
    def __init__(self, value_name, value_definition):
        self._value_name = value_name
        self._value_definition = value_definition
        self.container = self._get_container()

    def _get_container(self):
        uses_choices = 'choices' in self._value_definition
        uses_completer = 'completer' in self._value_definition
        if uses_choices:
            answer = WizardPromptSelectionAnswer(self._value_name)
        elif uses_completer:
            answer = WizardPromptCompletionAnswer(
                self._value_name,
                default_value=self._value_definition.get('default_value'),
                completer=self._value_definition['completer'],
            )
        else:
            answer = WizardPromptAnswer(
                self._value_name,
                default_value=self._value_definition.get('default_value'),
            )
        return ConditionalContainer(
            VSplit(
                [
                    WizardPromptDescription(
                        self._value_name, self._value_definition['description']
                    ),
                    answer,
                ],
            ),
            Condition(self._is_visible),
        )

    def _is_visible(self):
        return get_app().traverser.is_prompt_visible(self._value_name)

    def __pt_container__(self):
        return self.container


class WizardPromptDescription:
    def __init__(self, value_name, value_description):
        self._value_name = value_name
        self._value_description = value_description
        self.container = self._get_container()

    def _get_container(self):
        content = f'{self._value_description}:'
        buffer = Buffer(document=Document(content), read_only=True)
        return Window(
            content=BufferControl(buffer=buffer, focusable=False),
            style=self._get_style,
            width=Dimension.exact(len(content) + 1),
            dont_extend_height=True,
        )

    def _get_style(self):
        if get_app().traverser.get_current_prompt() == self._value_name:
            return 'class:wizard.prompt.description.current'
        else:
            return 'class:wizard.prompt.description'

    def __pt_container__(self):
        return self.container


class WizardPromptAnswer:
    def __init__(self, value_name, default_value=None):
        self._value_name = value_name
        self._default_value = ''
        if default_value is not None:
            self._default_value = str(default_value)
        self._buffer = self._get_answer_buffer()
        self.container = self._get_answer_container()

    def _get_answer_buffer(self):
        return Buffer(
            name=self._value_name, document=Document(text=self._default_value)
        )

    def _get_answer_container(self):
        return FullyExtendedWidthWindow(
            content=BufferControl(buffer=self._buffer),
            style=self._get_style,
            dont_extend_height=True,
        )

    def _get_style(self):
        if get_app().traverser.get_current_prompt() == self._value_name:
            return 'class:wizard.prompt.answer.current'
        else:
            return 'class:wizard.prompt.answer'

    def __pt_container__(self):
        return self.container


class WizardPromptCompletionAnswer(WizardPromptAnswer):
    COMPLETION_MENU_MAX_HEIGHT = 5

    def __init__(self, value_name, default_value=None, completer=None):
        self._completer = self._get_completer(completer)
        super().__init__(value_name, default_value)

    def _get_completer(self, completer):
        return {'file_completer': PathCompleter(expanduser=True)}[completer]

    def _get_answer_buffer(self):
        return Buffer(
            name=self._value_name,
            completer=self._completer,
            complete_while_typing=True,
            document=Document(text=self._default_value),
        )

    def _get_menu_height(self):
        if self._buffer.complete_state:
            return min(
                len(self._buffer.complete_state.completions),
                self.COMPLETION_MENU_MAX_HEIGHT,
            )
        return 0

    def _get_answer_container(self):
        return FloatContainer(
            FullyExtendedWidthWindow(
                content=BufferControl(buffer=self._buffer),
                style=self._get_style,
                wrap_lines=True,
            ),
            floats=[
                Float(
                    xcursor=True,
                    ycursor=True,
                    top=1,
                    height=self._get_menu_height,
                    content=CompletionsMenu(),
                )
            ],
            key_bindings=self._get_key_bindings(),
        )

    def _get_key_bindings(self):
        kb = KeyBindings()
        is_completing = Condition(
            lambda: getattr(self._buffer, 'complete_state', False)
        )

        @kb.add(Keys.Enter, filter=is_completing)
        def apply_completion(event):
            current_completion = self._buffer.complete_state.current_completion
            self._buffer.apply_completion(current_completion)

        @kb.add(Keys.BackTab, filter=is_completing)
        def apply_completion(event):
            self._buffer.complete_previous()

        @kb.add(Keys.Tab, filter=is_completing)
        def apply_completion(event):
            self._buffer.complete_next()

        return kb


class WizardPromptSelectionAnswer(WizardPromptAnswer):
    def _get_answer_container(self):
        return FullyExtendedWidthWindow(
            content=CollapsableSelectionMenuControl(
                items=self._get_choices,
                selection_capture_buffer=self._buffer,
                on_toggle=self._show_details,
            ),
            style=self._get_style,
            always_hide_cursor=True,
            scroll_offsets=ScrollOffsets(),
            right_margins=[ScrollbarMargin()],
        )

    def _get_choices(self):
        return get_app().traverser.get_current_prompt_choices()

    def _show_details(self, choice):
        app = get_app()
        details_buffer = app.layout.get_buffer_by_name('details_buffer')
        details, title = self._get_details(choice)
        app.details_title = title
        details_buffer.reset()
        new_document = Document(text=details, cursor_position=0)
        details_buffer.set_document(new_document, bypass_readonly=True)

    def _get_details(self, choice):
        app = get_app()
        details = ''
        title = get_app().traverser.get_details_title()
        if app.details_visible:
            details = get_app().traverser.get_details_for_choice(choice)
        return details, title



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile

from awscrt.crypto import ED25519, ED25519ExportFormat

from awscli.compat import compat_shell_quote
from awscli.customizations.commands import BasicCommand
from awscli.customizations.ec2instanceconnect.eicefetcher import (
    InstanceConnectEndpointRequestFetcher,
)
from awscli.customizations.exceptions import (
    ConfigurationError,
    ParamValidationError,
)

logger = logging.getLogger(__name__)


class SshCommand(BasicCommand):
    NAME = 'ssh'

    CONNECTION_TYPES = ['auto', 'direct', 'eice']

    ARG_TABLE = [
        {
            'name': 'instance-id',
            'help_text': 'Specify the ID of the EC2 instance to connect to.',
            'required': True,
        },
        {
            'name': 'instance-ip',
            'help_text': 'Specify the IP address of the instance to connect to.',
            'required': False,
        },
        {
            'name': 'private-key-file',
            'help_text': 'Specify the path to the private key file for login.',
            'required': False,
        },
        {
            'name': 'os-user',
            'help_text': 'Specify the OS SSH user. (Default: ec2-user)',
            'required': False,
            'default': 'ec2-user',
        },
        {
            'name': 'ssh-port',
            'help_text': 'Specify the SSH port. (Default: 22)',
            'required': False,
            'cli_type_name': 'integer',
            'default': '22',
        },
        {
            'name': 'local-forwarding',
            'help_text': 'Specify the local forwarding specification as defined by your OpenSSH client. '
            '(Example: 3336:remote.host:3306)',
            'required': False,
        },
        {
            'name': 'connection-type',
            'help_text': (
                'Specify how to SSH into the instance. (Default: auto)'
                '<ul>'
                '<li>direct: SSH directly to the instance. '
                'The CLI tries to connect using the IP addresses in the following order: '
                '<ul>'
                '<li>Public IPv4</li>'
                '<li>IPv6</li>'
                '<li>Private IPv4</li>'
                '</ul>'
                '</li>'
                '<li>eice: SSH using EC2 Instance Connect Endpoint. The CLI always uses the private IPv4 address.</li>'
                '<li>auto: The CLI automatically determines the connection type (direct or eice) '
                'to use based on the instance info. Currently the CLI tries to connect using the IP addresses '
                'in the following order and with the corresponding connection type:'
                '<ul>'
                '<li>Public IPv4: direct</li>'
                '<li>Private IPv4: eice</li>'
                '<li>IPv6: direct</li>'
                '</ul>'
                '</li>'
                '</ul>'
                'In the future, we might change the behavior of the auto connection type. To ensure that your '
                'desired connection type is used, we recommend that you explicitly set the --connection-type '
                'to either direct or eice.'
            ),
            'choices': CONNECTION_TYPES,
            'default': 'auto',
        },
        {
            'name': 'eice-options',
            'help_text': 'Specify the options for your EC2 Instance Connect Endpoint.',
            'schema': {
                'type': 'object',
                'properties': {
                    'maxTunnelDuration': {
                        'type': 'integer',
                        'description': (
                            'Specify the maximum duration (in seconds) for an established TCP connection. '
                            'Default: 3600 seconds (1 hour).'
                        ),
                        'required': False,
                    },
                    'endpointId': {
                        'type': 'string',
                        'description': (
                            'Specify the EC2 Instance Connect Endpoint ID to use to open the tunnel.'
                        ),
                        'required': False,
                    },
                    'dnsName': {
                        'type': 'string',
                        'description': (
                            'Specify the EC2 Instance Connect Endpoint DNS name to use to open '
                            'the tunnel. If this is specified, you must also specify endpointId.'
                        ),
                        'required': False,
                    },
                },
            },
        },
    ]

    DESCRIPTION = 'Connect to your EC2 instance using your OpenSSH client.'

    def __init__(self, session, key_manager=None):
        super().__init__(session)
        self._key_manager = (
            KeyManager() if (key_manager is None) else key_manager
        )

    def _run_main(self, parsed_args, parsed_globals):
        self._validate_parsed_args(parsed_args)

        # Describe instance to validate instance exist
        ec2_client = self._session.create_client(
            'ec2',
            region_name=parsed_globals.region,
            verify=parsed_globals.verify_ssl,
            endpoint_url=parsed_globals.endpoint_url,
        )
        instance_metadata = ec2_client.describe_instances(
            InstanceIds=[parsed_args.instance_id]
        )['Reservations'][0]['Instances'][0]

        private_ip_address = instance_metadata.get('PrivateIpAddress')
        public_ip_address = instance_metadata.get('PublicIpAddress')
        ipv6_address = instance_metadata.get('Ipv6Address')

        use_open_tunnel = False
        ip_address_to_connect = None
        if parsed_args.instance_ip:
            use_open_tunnel = parsed_args.connection_type == 'eice'
            ip_address_to_connect = parsed_args.instance_ip
        elif parsed_args.connection_type == 'eice' or parsed_args.eice_options:
            use_open_tunnel = True
            ip_address_to_connect = private_ip_address
        elif parsed_args.connection_type == 'direct':
            use_open_tunnel = False
            ip_address_to_connect = (
                public_ip_address or ipv6_address or private_ip_address
            )
        elif parsed_args.connection_type == 'auto':
            # In case of auto we use IPv4 address first and then IPv6 because right now most instance have these, but
            # in future we might want to switch this logic to where we select IPv6 first and then fallback to IPv4.
            if public_ip_address:
                use_open_tunnel = False
                ip_address_to_connect = public_ip_address
            elif private_ip_address:
                use_open_tunnel = True
                ip_address_to_connect = private_ip_address
            elif ipv6_address:
                use_open_tunnel = False
                ip_address_to_connect = ipv6_address

        if ip_address_to_connect is None:
            raise ParamValidationError(
                'Unable to find any IP address on the instance to connect to.'
            )

        instance_connect_endpoint_id = self._get_eice_option(
            parsed_args.eice_options, 'endpointId'
        )
        instance_connect_endpoint_dns_name = self._get_eice_option(
            parsed_args.eice_options, 'dnsName'
        )
        if use_open_tunnel and not instance_connect_endpoint_dns_name:
            eice_fetcher = InstanceConnectEndpointRequestFetcher()
            eice_info = eice_fetcher.get_available_instance_connect_endpoint(
                ec2_client,
                instance_metadata["VpcId"],
                instance_metadata["SubnetId"],
                instance_connect_endpoint_id,
            )
            instance_connect_endpoint_id = eice_info[
                "InstanceConnectEndpointId"
            ]

            is_fips_enabled = self._session.get_config_variable(
                'use_fips_endpoint'
            )
            instance_connect_endpoint_dns_name = (
                eice_fetcher.get_eice_dns_name(eice_info, is_fips_enabled)
            )

        # Validate ssh key exist (either use user defined or create new one)
        key_file = parsed_args.private_key_file

        with tempfile.TemporaryDirectory() as tmp_dir:
            if not key_file:
                key_file = os.path.join(tmp_dir, 'private-key')
                logger.debug('Generate new ssh key and upload')
                private_pem_bytes = self._generate_and_upload_ssh_key(
                    parsed_args.instance_id,
                    parsed_args.os_user,
                    parsed_globals,
                )

                with open(key_file, 'wb') as fd:
                    fd.write(private_pem_bytes)
                    logger.debug('Generated temporary key file: %s', key_file)
                os.chmod(key_file, 0o400)

            return self._ssh(
                use_open_tunnel,
                parsed_args.instance_id,
                parsed_args.ssh_port,
                parsed_args.os_user,
                parsed_args.local_forwarding,
                key_file,
                ip_address_to_connect,
                instance_connect_endpoint_id,
                instance_connect_endpoint_dns_name,
                self._get_eice_option(
                    parsed_args.eice_options, 'maxTunnelDuration'
                ),
                parsed_globals,
            )

    def _validate_parsed_args(self, parsed_args):
        if parsed_args.instance_id:
            if not re.search("^i-[0-9a-zA-Z]+$", parsed_args.instance_id):
                raise ParamValidationError(
                    'The specified instance ID is invalid. '
                    'Provide the full instance ID in the form i-xxxxxxxxxxxxxxxxx.'
                )

        eice_options = parsed_args.eice_options
        if parsed_args.connection_type == "direct" and eice_options:
            raise ParamValidationError(
                'eice-options can\'t be specified when connection type is direct.'
            )

        if self._get_eice_option(
            eice_options, 'dnsName'
        ) and not self._get_eice_option(eice_options, 'endpointId'):
            raise ParamValidationError(
                'When specifying dnsName, you must specify endpointId.'
            )

        if eice_options and 'maxTunnelDuration' in eice_options:
            max_tunnel_duration = eice_options['maxTunnelDuration']
            if max_tunnel_duration is not None and (
                max_tunnel_duration < 1 or max_tunnel_duration > 3_600
            ):
                raise ParamValidationError(
                    'Invalid value specified for maxTunnelDuration. Value must be greater than 1 and '
                    'less than 3600.'
                )

        if self._get_eice_option(eice_options, 'endpointId'):
            if not re.search(
                "^eice-[0-9a-zA-Z_]+$", eice_options['endpointId']
            ):
                raise ParamValidationError(
                    'The specified endpointId is invalid. '
                    'Provide the full EC2 Instance Connect Endpoint ID in '
                    'the form eice-xxxxxxxxxxxxxxxxx.'
                )

        if self._get_eice_option(eice_options, 'dnsName'):
            if not re.search('^[0-9a-zA-Z.-]+$', eice_options['dnsName']):
                raise ParamValidationError('The specified dnsName is invalid.')

        if parsed_args.instance_ip and parsed_args.connection_type == 'auto':
            raise ParamValidationError(
                'When specifying instance-ip, you must specify connection-type.'
            )

    def _get_eice_option(self, eice_options, option):
        if eice_options:
            return eice_options.get(option)
        return None

    def _generate_and_upload_ssh_key(
        self, instance_id, os_user, parsed_globals
    ):
        private_key = self._key_manager.generate_key()

        logger.debug('Upload public ssh key to instance')
        ec2_instance_connect_client = self._session.create_client(
            'ec2-instance-connect',
            region_name=parsed_globals.region,
            verify=parsed_globals.verify_ssl,
        )
        public_key = self._key_manager.get_public_key(private_key)
        self._key_manager.upload_public_key(
            ec2_instance_connect_client, instance_id, os_user, public_key
        )

        return self._key_manager.get_private_pem(private_key)

    def _generate_open_tunnel_command(
        self,
        instance_id,
        private_ip_address,
        ssh_port,
        eice_id,
        eice_dns_name,
        max_tunnel_duration,
        parsed_globals,
    ):
        aws_cli_path = sys.argv[0]
        proxy_command = [
            aws_cli_path,
            'ec2-instance-connect',
            'open-tunnel',
            '--instance-id',
            instance_id,
            '--private-ip-address',
            private_ip_address,
            '--remote-port',
            str(ssh_port),
        ]
        logger.debug(f"Using aws: {aws_cli_path}")

        if parsed_globals.region:
            proxy_command.append('--region')
            proxy_command.append(parsed_globals.region)
        if parsed_globals.profile:
            proxy_command.append('--profile')
            proxy_command.append(parsed_globals.profile)
        if eice_id:
            proxy_command.append('--instance-connect-endpoint-id')
            proxy_command.append(eice_id)
        if eice_dns_name:
            proxy_command.append('--instance-connect-endpoint-dns-name')
            proxy_command.append(eice_dns_name)
        if max_tunnel_duration:
            proxy_command.append('--max-tunnel-duration')
            proxy_command.append(str(max_tunnel_duration))

        return proxy_command

    def _ssh(
        self,
        use_open_tunnel,
        instance_id,
        ssh_port,
        os_user,
        local_forwarding,
        key_file,
        ip_address,
        eice_id,
        eice_dns_name,
        max_tunnel_duration,
        parsed_globals,
    ):
        proxy_command = self._generate_open_tunnel_command(
            instance_id,
            ip_address,
            ssh_port,
            eice_id,
            eice_dns_name,
            max_tunnel_duration,
            parsed_globals,
        )

        command = [
            'ssh',
            # adding ServerAliveInterval as default because it offers better customer experience as it let customer
            # know about terminated connections. If we want to allow customer to override this we can add additional
            # parameter to this cli command
            '-o',
            'ServerAliveInterval=5',
            '-p',
            str(ssh_port),
            '-i',
            key_file,
            os_user + '@' + ip_address,
        ]

        ssh_path = shutil.which('ssh')
        if ssh_path:
            command[0] = ssh_path
            logger.debug(f"Using ssh: {ssh_path}")
        else:
            raise ConfigurationError(
                'SSH not available. Please refer to the documentation '
                'at https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/Connect-using-EC2-Instance-Connect-Endpoint.html.'
            )

        # Add local-forwarding option if users passed local-forwarding
        if local_forwarding:
            command.insert(-1, '-L')
            command.insert(-1, local_forwarding)

        # If --debug is define lets add '-v' to ssh to generate debug level logs
        if parsed_globals.debug:
            command.insert(-1, '-v')

        # If we are trying to connect to instance in private subnet lets use open-tunnel command to use eice
        if use_open_tunnel:
            command.insert(-1, '-o')
            command.insert(
                -1,
                f"ProxyCommand={' '.join(compat_shell_quote(a) for a in proxy_command)}",
            )

        logger.debug('Invoking SSH command: %s', command)
        rc = subprocess.call(command)
        return rc


class KeyManager:
    def generate_key(self):
        logger.debug('Generate Ed25519 Key')
        return ED25519.new_generate()

    def _get_public_key(self, private_key):
        public_key_bytes = ED25519.export_public_key(
            private_key, export_format=ED25519ExportFormat.OPENSSH_B64
        )
        return f"ssh-ed25519 {public_key_bytes.decode()}"

    def get_public_key(self, private_key):
        return self._get_public_key(private_key)

    def _get_private_pem(self, private_key):
        private_key_bytes = ED25519.export_private_key(
            private_key, export_format=ED25519ExportFormat.OPENSSH_B64
        )
        return (
            '-----BEGIN OPENSSH PRIVATE KEY-----\n'
            f'{private_key_bytes.decode()}\n'
            '-----END OPENSSH PRIVATE KEY-----\n'
        ).encode()

    def get_private_pem(self, private_key):
        return self._get_private_pem(private_key=private_key)

    def upload_public_key(
        self, ec2_instance_connect_client, instance_id, os_user, public_key
    ):
        logger.debug('Upload ssh key to instance')
        ec2_instance_connect_client.send_ssh_public_key(
            InstanceId=instance_id,
            InstanceOSUser=os_user,
            SSHPublicKey=public_key,
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# Declare all the constants used by DataPipeline in this file

# DataPipeline role names
DATAPIPELINE_DEFAULT_SERVICE_ROLE_NAME = "DataPipelineDefaultRole"
DATAPIPELINE_DEFAULT_RESOURCE_ROLE_NAME = "DataPipelineDefaultResourceRole"

# DataPipeline role arn names
DATAPIPELINE_DEFAULT_SERVICE_ROLE_ARN = (
    "arn:aws:iam::aws:policy/" "service-role/AWSDataPipelineRole"
)
DATAPIPELINE_DEFAULT_RESOURCE_ROLE_ARN = (
    "arn:aws:iam::aws:policy/"
    "service-role/"
    "AmazonEC2RoleforDataPipelineRole"
)

# Assume Role Policy definitions for roles
DATAPIPELINE_DEFAULT_RESOURCE_ROLE_ASSUME_POLICY = {
    "Version": "2008-10-17",
    "Statement": [
        {
            "Sid": "",
            "Effect": "Allow",
            "Principal": {"Service": "ec2.amazonaws.com"},
            "Action": "sts:AssumeRole",
        }
    ],
}

DATAPIPELINE_DEFAULT_SERVICE_ROLE_ASSUME_POLICY = {
    "Version": "2008-10-17",
    "Statement": [
        {
            "Sid": "",
            "Effect": "Allow",
            "Principal": {
                "Service": [
                    "datapipeline.amazonaws.com",
                    "elasticmapreduce.amazonaws.com",
                ]
            },
            "Action": "sts:AssumeRole",
        }
    ],
}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io
import os
import sqlite3
import sys
import threading
import time
import uuid
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

from botocore.compat import get_md5
from botocore.exceptions import MD5UnavailableError
from botocore.useragent import UserAgentComponent

from awscli.compat import is_windows
from awscli.utils import add_component_to_user_agent_extra

_CACHE_DIR = Path.home() / '.aws' / 'cli' / 'cache'
_DATABASE_FILENAME = 'session.db'
_SESSION_LENGTH_SECONDS = 60 * 30
_SESSION_ID_LENGTH = 12


def _get_checksum():
    hashlib_params = {"usedforsecurity": False}
    try:
        checksum = get_md5(**hashlib_params)
    except MD5UnavailableError:
        import hashlib

        checksum = hashlib.sha256(**hashlib_params)
    return checksum


@dataclass
class CLISessionData:
    key: str
    session_id: str
    timestamp: int


class CLISessionDatabaseConnection:
    _CREATE_TABLE = """
        CREATE TABLE IF NOT EXISTS session (
          key TEXT PRIMARY KEY,
          session_id TEXT NOT NULL,
          timestamp INTEGER NOT NULL
        )
    """
    _CREATE_HOST_ID_TABLE = """
        CREATE TABLE IF NOT EXISTS host_id (
          key INTEGER PRIMARY KEY,
          id TEXT UNIQUE NOT NULL
        )
    """
    _CHECK_HOST_ID = """
        SELECT COUNT(*) FROM host_id
    """
    _INSERT_HOST_ID = """
        INSERT OR IGNORE INTO host_id (
            key, id
        ) VALUES (?, ?)
    """
    _ENABLE_WAL = 'PRAGMA journal_mode=WAL'

    def __init__(self, connection=None):
        self._connection = connection or sqlite3.connect(
            _CACHE_DIR / _DATABASE_FILENAME,
            check_same_thread=False,
            isolation_level=None,
        )
        self._ensure_cache_dir()
        self._ensure_database_setup()

    def execute(self, query, *parameters):
        try:
            return self._connection.execute(query, *parameters)
        except sqlite3.OperationalError:
            # Process timed out waiting for database lock.
            # Return any empty `Cursor` object instead of
            # raising an exception.
            return sqlite3.Cursor(self._connection)

    def _ensure_cache_dir(self):
        _CACHE_DIR.mkdir(parents=True, exist_ok=True)

    def _ensure_database_setup(self):
        self._create_session_table()
        self._create_host_id_table()
        self._ensure_host_id()
        self._try_to_enable_wal()

    def _create_session_table(self):
        self.execute(self._CREATE_TABLE)

    def _create_host_id_table(self):
        self.execute(self._CREATE_HOST_ID_TABLE)

    def _ensure_host_id(self):
        cur = self.execute(self._CHECK_HOST_ID)
        host_id_ct = cur.fetchone()[0]
        if host_id_ct == 0:
            self.execute(
                self._INSERT_HOST_ID,
                # Hardcode `0` as primary key to ensure
                # there's only ever 1 host id in the table.
                (
                    0,
                    str(uuid.uuid4()),
                ),
            )

    def _try_to_enable_wal(self):
        try:
            self.execute(self._ENABLE_WAL)
        except sqlite3.Error:
            # This is just a performance enhancement so it is optional. Not all
            # systems will have a sqlite compiled with the WAL enabled.
            pass


class CLISessionDatabaseWriter:
    _WRITE_RECORD = """
        INSERT OR REPLACE INTO session (
            key, session_id, timestamp
        ) VALUES (?, ?, ?)
    """

    def __init__(self, connection):
        self._connection = connection

    def write(self, data):
        self._connection.execute(
            self._WRITE_RECORD,
            (
                data.key,
                data.session_id,
                data.timestamp,
            ),
        )


class CLISessionDatabaseReader:
    _READ_RECORD = """
        SELECT *
        FROM session
        WHERE key = ?
    """
    _READ_HOST_ID = """
        SELECT id
        FROM host_id
        WHERE key = 0
    """

    def __init__(self, connection):
        self._connection = connection

    def read(self, key):
        cursor = self._connection.execute(self._READ_RECORD, (key,))
        result = cursor.fetchone()
        if result is None:
            return
        return CLISessionData(*result)

    def read_host_id(self):
        cursor = self._connection.execute(self._READ_HOST_ID)
        return cursor.fetchone()[0]


class CLISessionDatabaseSweeper:
    _DELETE_RECORDS = """
        DELETE FROM session
        WHERE timestamp < ?
    """

    def __init__(self, connection):
        self._connection = connection

    def sweep(self, timestamp):
        try:
            self._connection.execute(self._DELETE_RECORDS, (timestamp,))
        except Exception:
            # This is just a background cleanup task. No need to
            # handle it or direct to stderr.
            return


class CLISessionGenerator:
    def generate_session_id(self, host_id, tty, timestamp):
        return self._generate_checksum(host_id, tty, timestamp)

    def generate_cache_key(self, host_id, tty):
        return self._generate_checksum(host_id, tty)

    def _generate_checksum(self, *args):
        checksum = _get_checksum()
        str_to_hash = ""
        for arg in args:
            if arg is not None:
                str_to_hash += str(arg)
        checksum.update(str_to_hash.encode('utf-8'))
        return checksum.hexdigest()[:_SESSION_ID_LENGTH]


class CLISessionOrchestrator:
    def __init__(self, generator, writer, reader, sweeper):
        self._generator = generator
        self._writer = writer
        self._reader = reader
        self._sweeper = sweeper

        self._sweep_cache()

    @cached_property
    def cache_key(self):
        return self._generator.generate_cache_key(self._host_id, self._tty)

    @cached_property
    def _session_id(self):
        return self._generator.generate_session_id(
            self._host_id, self._tty, self._timestamp
        )

    @cached_property
    def session_id(self):
        if (cached_data := self._reader.read(self.cache_key)) is not None:
            # Cache hit, but session id is expired. Generate new id and update.
            if (
                cached_data.timestamp + _SESSION_LENGTH_SECONDS
                < self._timestamp
            ):
                cached_data.session_id = self._session_id
            # Always update the timestamp to last used.
            cached_data.timestamp = self._timestamp
            self._writer.write(cached_data)
            return cached_data.session_id
        # Cache miss, generate and write new record.
        session_id = self._session_id
        session_data = CLISessionData(
            self.cache_key, session_id, self._timestamp
        )
        self._writer.write(session_data)
        return session_id

    @cached_property
    def _tty(self):
        # os.ttyname is only available on Unix platforms.
        if is_windows:
            return
        try:
            return os.ttyname(sys.stdin.fileno())
        except (OSError, io.UnsupportedOperation):
            # Standard input was redirected to a pseudofile.
            # This can happen when running tests on IDEs or
            # running scripts with redirected input, etc.
            return

    @cached_property
    def _host_id(self):
        return self._reader.read_host_id()

    @cached_property
    def _timestamp(self):
        return int(time.time())

    def _sweep_cache(self):
        t = threading.Thread(
            target=self._sweeper.sweep,
            args=(self._timestamp - _SESSION_LENGTH_SECONDS,),
            daemon=True,
        )
        t.start()


def _get_cli_session_orchestrator():
    conn = CLISessionDatabaseConnection()
    return CLISessionOrchestrator(
        CLISessionGenerator(),
        CLISessionDatabaseWriter(conn),
        CLISessionDatabaseReader(conn),
        CLISessionDatabaseSweeper(conn),
    )


def add_session_id_component_to_user_agent_extra(session, orchestrator=None):
    try:
        cli_session_orchestrator = (
            orchestrator or _get_cli_session_orchestrator()
        )
        add_component_to_user_agent_extra(
            session,
            UserAgentComponent("sid", cli_session_orchestrator.session_id),
        )
    except Exception:
        # Ideally, the AWS CLI should never throw if the session id
        # can't be generated since it's not critical for users. Issues
        # with session data should instead be caught server-side.
        pass



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at

#     http://aws.amazon.com/apache2.0/

# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
from datetime import datetime

from botocore.compat import json
from botocore.paginate import PageIterator
from botocore.utils import set_value_from_jmespath
from ruamel.yaml import YAML

from awscli import compat, text
from awscli.table import ColorizedStyler, MultiTable, Styler
from awscli.utils import json_encoder

LOG = logging.getLogger(__name__)


def is_response_paginated(response):
    return isinstance(response, PageIterator)


class Formatter:
    def __init__(self, args):
        self._args = args

    def _get_transformed_response_for_output(self, response):
        # Performs various transformations to the API response ready to
        # be outputted such as removing response id's and performing the
        # query
        self._remove_request_id(response)
        return self._apply_query_if_needed(response)

    def _remove_request_id(self, response_data):
        if 'ResponseMetadata' in response_data:
            if 'RequestId' in response_data['ResponseMetadata']:
                request_id = response_data['ResponseMetadata']['RequestId']
                LOG.debug('RequestId: %s', request_id)
            del response_data['ResponseMetadata']

    def _apply_query_if_needed(self, response_data):
        if self._args.query is not None:
            response_data = self._args.query.search(response_data)
        return response_data

    def _get_default_stream(self):
        return compat.get_stdout_text_writer()

    def _flush_stream(self, stream):
        try:
            stream.flush()
        except OSError:
            pass


class FullyBufferedFormatter(Formatter):
    def __call__(self, command_name, response, stream=None):
        if stream is None:
            # Retrieve stdout on invocation instead of at import time
            # so that if anything wraps stdout we'll pick up those changes
            # (specifically colorama on windows wraps stdout).
            stream = self._get_default_stream()
        compat.set_preferred_output_encoding(stream)
        # I think the interfaces between non-paginated
        # and paginated responses can still be cleaned up.
        if is_response_paginated(response):
            response_data = response.build_full_result()
        else:
            response_data = response
        response_data = self._get_transformed_response_for_output(
            response_data
        )
        try:
            self._format_response(command_name, response_data, stream)
        except OSError:
            # If the reading end of our stdout stream has closed the file
            # we can just exit.
            pass
        finally:
            # flush is needed to avoid the "close failed in file object
            # destructor" in python2.x (see http://bugs.python.org/issue11380).
            self._flush_stream(stream)


class JSONFormatter(FullyBufferedFormatter):
    def _format_response(self, command_name, response, stream):
        # For operations that have no response body (e.g. s3 put-object)
        # the response will be an empty string.  We don't want to print
        # that out to the user but other "falsey" values like an empty
        # dictionary should be printed.
        if response != {}:
            json.dump(
                response,
                stream,
                indent=4,
                default=json_encoder,
                ensure_ascii=False,
            )
            stream.write('\n')


class YAMLDumper:
    def __init__(self):
        self._yaml = YAML(typ='safe')
        # Encoding is set to None because we handle the encoding by
        # wrapping the stream, so there's no need for the yaml library
        # to do it.
        self._yaml.encoding = None
        self._yaml.representer.default_flow_style = False

    def dump(self, value, stream):
        if self._is_json_scalar(value) or isinstance(value, datetime):
            # YAML will attempt to disambiguate scalars by ending the stream
            # with an elipsis. While this is technically valid YAML,
            # it's not particularly useful. Unfortunately there's no
            # universal way around this, so instead we just json dump the
            # values. Also note that datetimes are explicitly not supported
            # - the json dumper will complain if you pass them in. datetime
            # values should respect the cli timestamp format, which is
            # impossible to do from the Formatter.
            json.dump(value, stream, ensure_ascii=False, default=json_encoder)
            stream.write('\n')
        else:
            self._yaml.dump(value, stream)

    def _is_json_scalar(self, value):
        if value is None:
            return True
        return isinstance(value, (int, float, bool, str))


class YAMLFormatter(FullyBufferedFormatter):
    def __init__(self, args, yaml_dumper=None):
        super(YAMLFormatter, self).__init__(args)
        self._yaml_dumper = yaml_dumper
        if yaml_dumper is None:
            self._yaml_dumper = YAMLDumper()

    def _format_response(self, command_name, response, stream):
        if response == {}:
            return None
        self._yaml_dumper.dump(response, stream)


class StreamedYAMLFormatter(Formatter):
    def __init__(self, args, yaml_dumper=None):
        super(StreamedYAMLFormatter, self).__init__(args)
        self._yaml_dumper = yaml_dumper
        if yaml_dumper is None:
            self._yaml_dumper = YAMLDumper()

    def __call__(self, command_name, response, stream=None):
        if stream is None:
            stream = self._get_default_stream()
        compat.set_preferred_output_encoding(stream)
        response_stream = self._get_response_stream(response)
        for response in response_stream:
            try:
                # For YAML it is ambiguous as to whether the output from the
                # stream is N responses in 1 list or N lists each with 1
                # response. We go with the latter so we can reuse our YAML
                # dumper
                self._yaml_dumper.dump([response], stream)
            except OSError:
                # If the reading end of our stdout stream has closed the file
                # we can just exit.
                return
            finally:
                # flush is needed to avoid the "close failed in file object
                # destructor" in python2.x. See:
                # http://bugs.python.org/issue11380).
                self._flush_stream(stream)

    def _get_response_stream(self, response):
        if is_response_paginated(response):
            return compat.imap(
                self._get_transformed_response_for_output, response
            )
        else:
            output = self._get_transformed_response_for_output(response)
            if output == {}:
                # The operation did not have an output so return an empty list
                # as the stream so nothing gets printed out.
                return []
            return [output]


class TableFormatter(FullyBufferedFormatter):
    """Pretty print a table from a given response.

    The table formatter is able to take any generic response
    and generate a pretty printed table.  It does this without
    using the output definition from the model.

    """

    def __init__(self, args, table=None):
        super(TableFormatter, self).__init__(args)
        if args.color == 'auto':
            self.table = MultiTable(
                initial_section=False, column_separator='|'
            )
        elif args.color == 'off':
            styler = Styler()
            self.table = MultiTable(
                initial_section=False, column_separator='|', styler=styler
            )
        elif args.color == 'on':
            styler = ColorizedStyler()
            self.table = MultiTable(
                initial_section=False, column_separator='|', styler=styler
            )
        else:
            raise ValueError("Unknown color option: %s" % args.color)

    def _format_response(self, command_name, response, stream):
        if self._build_table(command_name, response):
            try:
                self.table.render(stream)
            except OSError:
                # If they're piping stdout to another process which exits
                # before we're done writing all of our output, we'll get an
                # error about a closed pipe which we can safely ignore.
                pass

    def _build_table(self, title, current, indent_level=0):
        if not current:
            return False
        if title is not None:
            self.table.new_section(title, indent_level=indent_level)
        if isinstance(current, list):
            if isinstance(current[0], dict):
                self._build_sub_table_from_list(current, indent_level, title)
            else:
                for item in current:
                    if self._scalar_type(item):
                        self.table.add_row([item])
                    elif all(self._scalar_type(el) for el in item):
                        self.table.add_row(item)
                    else:
                        self._build_table(title=None, current=item)
        if isinstance(current, dict):
            # Render a single row section with keys as header
            # and the row as the values, unless the value
            # is a list.
            self._build_sub_table_from_dict(current, indent_level)
        return True

    def _build_sub_table_from_dict(self, current, indent_level):
        # Render a single row section with keys as header
        # and the row as the values, unless the value
        # is a list.
        headers, more = self._group_scalar_keys(current)
        if len(headers) == 1:
            # Special casing if a dict has a single scalar key/value pair.
            self.table.add_row([headers[0], current[headers[0]]])
        elif headers:
            self.table.add_row_header(headers)
            self.table.add_row([current[k] for k in headers])
        for remaining in more:
            self._build_table(
                remaining, current[remaining], indent_level=indent_level + 1
            )

    def _build_sub_table_from_list(self, current, indent_level, title):
        headers, more = self._group_scalar_keys_from_list(current)
        self.table.add_row_header(headers)
        first = True
        for element in current:
            if not first and more:
                self.table.new_section(title, indent_level=indent_level)
                self.table.add_row_header(headers)
            first = False
            # Use .get() to account for the fact that sometimes an element
            # may not have all the keys from the header.
            self.table.add_row([element.get(header, '') for header in headers])
            for remaining in more:
                # Some of the non scalar attributes may not necessarily
                # be in every single element of the list, so we need to
                # check this condition before recursing.
                if remaining in element:
                    self._build_table(
                        remaining,
                        element[remaining],
                        indent_level=indent_level + 1,
                    )

    def _scalar_type(self, element):
        return not isinstance(element, (list, dict))

    def _group_scalar_keys_from_list(self, list_of_dicts):
        # We want to make sure we catch all the keys in the list of dicts.
        # Most of the time each list element has the same keys, but sometimes
        # a list element will have keys not defined in other elements.
        headers = set()
        more = set()
        for item in list_of_dicts:
            current_headers, current_more = self._group_scalar_keys(item)
            headers.update(current_headers)
            more.update(current_more)
        headers = list(sorted(headers))
        more = list(sorted(more))
        return headers, more

    def _group_scalar_keys(self, current):
        # Given a dict, separate the keys into those whose values are
        # scalar, and those whose values aren't.  Return two lists,
        # one is the scalar value keys, the second is the remaining keys.
        more = []
        headers = []
        for element in current:
            if self._scalar_type(current[element]):
                headers.append(element)
            else:
                more.append(element)
        headers.sort()
        more.sort()
        return headers, more


class TextFormatter(Formatter):
    def __call__(self, command_name, response, stream=None):
        if stream is None:
            stream = self._get_default_stream()
        try:
            if is_response_paginated(response):
                result_keys = response.result_keys
                for i, page in enumerate(response):
                    if i > 0:
                        current = {}
                    else:
                        current = response.non_aggregate_part

                    for result_key in result_keys:
                        data = result_key.search(page)
                        set_value_from_jmespath(
                            current, result_key.expression, data
                        )
                    self._format_response(current, stream)
                if response.resume_token:
                    # Tell the user about the next token so they can continue
                    # if they want.
                    self._format_response(
                        {'NextToken': {'NextToken': response.resume_token}},
                        stream,
                    )
            else:
                self._remove_request_id(response)
                self._format_response(response, stream)
        finally:
            # flush is needed to avoid the "close failed in file object
            # destructor" in python2.x (see http://bugs.python.org/issue11380).
            self._flush_stream(stream)

    def _format_response(self, response, stream):
        if self._args.query is not None:
            expression = self._args.query
            response = expression.search(response)
        text.format_text(response, stream)


class OffFormatter(Formatter):
    """Formatter that suppresses all output.
    Only stdout is suppressed; stderr (error messages) remains visible.
    """

    def __call__(self, command_name, response, stream=None):
        if is_response_paginated(response):
            for _ in response:
                pass


CLI_OUTPUT_FORMATS = {
    'json': JSONFormatter,
    'text': TextFormatter,
    'table': TableFormatter,
    'yaml': YAMLFormatter,
    'yaml-stream': StreamedYAMLFormatter,
    'off': OffFormatter,
}


def get_formatter(format_type, args):
    if format_type not in CLI_OUTPUT_FORMATS:
        raise ValueError("Unknown output type: %s" % format_type)
    format_type_cls = CLI_OUTPUT_FORMATS[format_type]
    return format_type_cls(args)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re
from collections import namedtuple


def fuzzy_filter(prefix, completions):
    # To filter we transform string to regex in such a way:
    #
    # "rmt" -> /r.*?m.*?t/
    #
    # and search for matching in all the strings. All the matches we
    # sort in such a way that shortest and the the closest to the beginning
    # match part coming first. It means that the first completion will
    # be a string where distance between matching letters as small
    # as possible and the first letter is as close to the beginning as possible

    if prefix and completions:
        fuzzy_matches = []
        # lookahead regex to manage overlapping matches
        pattern = ".*?".join(map(re.escape, prefix))
        pattern = f'(?=({pattern}))'
        regex = re.compile(pattern, re.IGNORECASE)
        for completion in completions:
            name = completion.display_text or completion.name
            matches = list(regex.finditer(name))
            if matches:
                # Prefer the match, closest to the left, then shortest.
                best = min(matches, key=lambda m: (m.start(), len(m.group(1))))
                fuzzy_matches.append(
                    _FuzzyMatch(len(best.group(1)), best.start(), completion)
                )
        return [
            x
            for _, _, x in sorted(
                fuzzy_matches,
                key=lambda match: (
                    match.match_length,
                    match.start_pos,
                    match.completion.display_text,
                    match.completion.name,
                ),
            )
        ]
    return completions


def startswith_filter(prefix, completions):
    return [
        completion
        for completion in completions
        if (completion.display_text or completion.name).startswith(prefix)
    ]


_FuzzyMatch = namedtuple('_FuzzyMatch', 'match_length start_pos completion')



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import select
import socket
import struct
import sys
import threading
import time
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from threading import Event
from urllib.parse import urlparse

from awscrt import websocket
from awscrt.http import HttpProxyAuthenticationType, HttpProxyOptions
from awscrt.io import ClientTlsContext, TlsContextOptions
from awscrt.websocket import (
    OnConnectionSetupData,
    OnConnectionShutdownData,
    OnIncomingFrameCompleteData,
    OnIncomingFramePayloadData,
    OnSendFrameCompleteData,
    Opcode,
)

from awscli.compat import is_windows

logger = logging.getLogger(__name__)


class WebsocketException(RuntimeError):
    pass


class InputClosedError(RuntimeError):
    pass


class BaseWebsocketIO:
    def has_data_to_read(self):
        raise NotImplementedError("has_data_to_read")

    def read(self, amt) -> bytes:
        raise NotImplementedError("read")

    def write(self, data):
        raise NotImplementedError("write")

    def close(self):
        raise NotImplementedError("close")


_SELECT_TIMEOUT = 0.05


class StdinStdoutIO(BaseWebsocketIO):
    def has_data_to_read(self):
        socket_list = [sys.stdin]
        read_sockets, _, _ = select.select(
            socket_list, [], [], _SELECT_TIMEOUT
        )
        if read_sockets:
            return True
        return False

    def read(self, amt) -> bytes:
        return sys.stdin.buffer.read1(amt)

    def write(self, data):
        sys.stdout.buffer.write(data)
        sys.stdout.flush()

    def close(self):
        pass


class WindowsStdinStdoutIO(StdinStdoutIO):
    def has_data_to_read(self):
        return True


class TCPSocketIO(BaseWebsocketIO):
    def __init__(self, conn):
        self.conn = conn

    def has_data_to_read(self):
        return True

    def read(self, amt) -> bytes:
        data = self.conn.recv(amt)
        # In listener mode use can CTRL+C during host verification that kills the client TCP connect,
        # when this happens we are able to successfully disconnect because has_data_to_read always return true.
        # This will check if data is empty and if yes then raise InputCloseError
        #
        # recv() relies on the underlying system call which returns empty bytes when the connection is closed.
        # Linux: https://manpages.debian.org/bullseye/manpages-dev/recv.2.en.html
        # Windows: https://learn.microsoft.com/en-us/windows/win32/api/winsock/nf-winsock-recv
        if not data:
            raise InputClosedError()
        return data

    def write(self, data):
        self.conn.sendall(data)

    def close(self):
        try:
            self.conn.close()
        # On Windows, we could receive an OSError if the tcp conn is already closed.
        except OSError:
            pass


class Websocket:
    def __init__(
        self,
        websocketio,
        websocket_id,
        tls_connection_options=None,
        on_connection_event=None,
        shutdown_event=None,
        send_frame_results_queue=None,
    ):
        self.websocketio = websocketio
        self.websocket_id = websocket_id
        self._on_connection_event = (
            Event() if (on_connection_event is None) else on_connection_event
        )
        self._send_frame_results_queue = (
            Queue()
            if (send_frame_results_queue is None)
            else send_frame_results_queue
        )
        self._shutdown_event = (
            Event() if (shutdown_event is None) else shutdown_event
        )
        self._websocket = None
        self._exception = None
        self._close_frame_bytes = bytearray()

        if tls_connection_options is None:
            self._tls_connection_options = ClientTlsContext(
                TlsContextOptions()
            ).new_connection_options()
        else:
            self._tls_connection_options = tls_connection_options

    _MAX_BYTES_PER_FRAME = 65_000
    _WAIT_INTERVAL_FOR_INPUT = 0.05

    @property
    def exception(self):
        return self._exception

    def connect(self, url, user_agent=None):
        parsed_url = urlparse(url)
        path = parsed_url.path + "?" + parsed_url.query
        request = websocket.create_handshake_request(
            host=parsed_url.hostname, path=path
        )
        if user_agent:
            request.headers.set("User-Agent", user_agent)

        environment = os.environ.copy()
        proxy_options = None
        proxy_url = environment.get("HTTP_PROXY") or environment.get(
            "HTTPS_PROXY"
        )
        no_proxy = environment.get("NO_PROXY", "")
        if proxy_url and parsed_url.hostname not in no_proxy:
            parsed_proxy_url = urlparse(proxy_url)
            logger.debug(
                f"Using the following proxy: {parsed_proxy_url.hostname}"
            )
            proxy_options = HttpProxyOptions(
                host_name=parsed_proxy_url.hostname,
                port=parsed_proxy_url.port,
                auth_type=HttpProxyAuthenticationType.Basic
                if proxy_url
                else HttpProxyAuthenticationType.Nothing,
                auth_username=parsed_proxy_url.username or None,
                auth_password=parsed_proxy_url.password or None,
            )

        self._tls_connection_options.set_server_name(parsed_url.hostname)
        websocket.connect(
            host=parsed_url.hostname,
            handshake_request=request,
            port=443,
            tls_connection_options=self._tls_connection_options,
            proxy_options=proxy_options,
            on_connection_setup=self._on_connection,
            on_connection_shutdown=self._on_connection_shutdown,
            on_incoming_frame_payload=self._on_incoming_frame_payload_data,
            on_incoming_frame_complete=self._on_incoming_frame_complete,
        )

        # Wait for the on_connection callback to be called.
        self._on_connection_event.wait()

        if not self._websocket:
            self.close()
            raise self._exception

    def write_data_from_input(self):
        try:
            # Start writing data to the websocket connection and block current thread.
            self._write_data_from_input()
        finally:
            self.close()

        if self._exception:
            raise self._exception

    def close(self):
        if self._websocket:
            self._websocket.close()
        self.websocketio.close()
        self._shutdown_event.set()

    def _write_data_from_input(self):
        while not self._shutdown_event.is_set():
            # Wait until there's some data to read
            if not self.websocketio.has_data_to_read():
                time.sleep(self._WAIT_INTERVAL_FOR_INPUT)
                continue

            try:
                data = self.websocketio.read(self._MAX_BYTES_PER_FRAME)
            except InputClosedError:
                logger.debug('Input closed. Shutting down websocket.')
                self.close()

            try:
                self._websocket.send_frame(
                    opcode=Opcode.BINARY,
                    payload=data,
                    on_complete=self._on_send_frame_complete_data,
                )
                # Block until send_frame on_complete
                self._send_frame_results_queue.get()
            except RuntimeError as e:
                crt_exceptions = [
                    "AWS_ERROR_HTTP_WEBSOCKET_CLOSE_FRAME_SENT",
                    "AWS_ERROR_HTTP_CONNECTION_CLOSED",
                ]
                # Expected exception if server or user closes conn. Catch it and gracefully exit this method.
                if any(exc_code in str(e.args) for exc_code in crt_exceptions):
                    logger.debug(
                        f"Received exception when sending websocket frame: {e.args}"
                    )
                    self.close()
                else:
                    raise e

    def _on_connection(self, data: OnConnectionSetupData) -> None:
        request_id_header = [
            header
            for header in data.handshake_response_headers
            if any("x-amzn-requestid" == key.lower() for key in header)
        ]
        if request_id_header:
            logger.debug(f"OpenTunnel RequestId: {request_id_header[0][1]}")
        if data.exception:
            if data.handshake_response_status:
                logger.debug(f"Status code: {data.handshake_response_status}")
            if data.handshake_response_body:
                decoded_response = data.handshake_response_body.decode("utf-8")
                logger.error(decoded_response)
            self._exception = data.exception
        self._websocket = data.websocket
        self._on_connection_event.set()

    def _on_incoming_frame_payload_data(
        self, incoming_frame_data: OnIncomingFramePayloadData
    ) -> None:
        opcode = incoming_frame_data.frame.opcode
        if not opcode.is_data_frame():
            if opcode == Opcode.CLOSE and incoming_frame_data.data:
                self._close_frame_bytes += incoming_frame_data.data
            return
        if opcode == Opcode.TEXT:
            self._exception = WebsocketException(
                "Received invalid data from server, closing websocket connection."
            )
            self.close()
            return
        self.websocketio.write(incoming_frame_data.data)

    def _on_incoming_frame_complete(
        self, incoming_frame_data: OnIncomingFrameCompleteData
    ):
        if (
            incoming_frame_data.frame.opcode == Opcode.CLOSE
            and incoming_frame_data.exception is None
        ):
            if len(self._close_frame_bytes) > 0:
                shutdown_code_as_bytes = self._close_frame_bytes[0:2]
                # The shutdown code is a packed 2 byte unsigned int.
                shutdown_code = struct.unpack(">H", shutdown_code_as_bytes)[0]
                shutdown_reason_in_bytes = self._close_frame_bytes[2:]
                if shutdown_code != 1000:
                    logger.debug("Shutdown code: %s", str(shutdown_code))
                    decoded_shutdown_reason = shutdown_reason_in_bytes.decode(
                        "utf-8"
                    )
                    self._exception = WebsocketException(
                        f"Websocket Closure Reason: {decoded_shutdown_reason}"
                    )
            self.close()

    def _on_send_frame_complete_data(
        self, send_frame_data: OnSendFrameCompleteData
    ) -> None:
        self._send_frame_results_queue.put(send_frame_data)

    def _on_connection_shutdown(self, data: OnConnectionShutdownData) -> None:
        if data.exception:
            self._exception = data.exception
        self.close()


class WebsocketManager:
    def __init__(
        self,
        port,
        max_websocket_connections,
        eice_request_signer,
        user_agent=None,
    ):
        self._port = port
        self._executor = ThreadPoolExecutor(
            max_workers=max_websocket_connections
        )
        self._connection_id_counter = 1
        self._max_websocket_connections = max_websocket_connections
        self._socket = None
        self._inflight_futures_and_websockets = []
        self._eice_request_signer = eice_request_signer
        self._user_agent = user_agent

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for _, web_socket in self._inflight_futures_and_websockets:
            # Close the websocket handlers.
            web_socket.close()
        if self._socket:
            try:
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
            # On Windows, if the socket is already closed, we will get an OSError.
            except OSError:
                pass
        self._executor.shutdown()

    # Used to break out of while loop in tests.
    RUNNING = threading.Event()

    def run(self):
        # If no port is specified, open a singular websocket connection.
        if not self._port:
            websocketio = (
                WindowsStdinStdoutIO() if is_windows else StdinStdoutIO()
            )
            future = self._open_websocket_connection(
                Websocket(websocketio, websocket_id=None)
            )
            # Block until the future completes.
            future.result()
        else:
            self._listen_on_port()

    def _listen_on_port(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.bind(("localhost", self._port))
        self._socket.listen()
        print(f"Listening for connections on port {self._port}.")
        while not self.RUNNING.is_set():
            socket_list = [self._socket]
            read_sockets, _, _ = select.select(socket_list, [], [], 0.1)
            if read_sockets:
                conn, addr = self._socket.accept()
                # Check if we have reached max connections
                self._remove_done_futures()
                if (
                    len(self._inflight_futures_and_websockets)
                    >= self._max_websocket_connections
                ):
                    print(
                        f"Max websocket connections {self._max_websocket_connections} have been reached, closing "
                        f"incoming connection."
                    )
                    conn.close()
                    continue

                websocket_id = self._connection_id_counter
                self._connection_id_counter += 1
                print(
                    f"[{websocket_id}] Accepted new tcp connection, opening websocket tunnel."
                )
                try:
                    web_socket = Websocket(TCPSocketIO(conn), websocket_id)
                    future = self._open_websocket_connection(web_socket)
                    future.add_done_callback(
                        self._print_tcp_conn_closed(web_socket)
                    )
                except WebsocketException as e:
                    logger.error(
                        f"[{websocket_id}] Encountered error opening websocket: {e.args}"
                    )

    def _open_websocket_connection(self, web_socket):
        presigned_url = self._eice_request_signer.get_presigned_url()
        web_socket.connect(presigned_url, self._user_agent)

        future = self._executor.submit(web_socket.write_data_from_input)

        self._inflight_futures_and_websockets.append((future, web_socket))
        return future

    def _print_tcp_conn_closed(self, web_socket):
        def _on_done_callback(future):
            if future.exception():
                logger.error(
                    f"[{web_socket.websocket_id}] Encountered error with websocket: {future.exception().args}"
                )
            print(f"[{web_socket.websocket_id}] Closing tcp connection.")

        return _on_done_callback

    def _remove_done_futures(self):
        self._inflight_futures_and_websockets = [
            (future, web_socket)
            for future, web_socket in self._inflight_futures_and_websockets
            if not future.done()
        ]



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class Base36:
    def str_to_int(self, request):
        """Method to convert given string into decimal representation"""
        result = 0
        for char in request:
            result = result * 256 + ord(char)

        return result

    def encode(self, request):
        """Method to return base36 encoded form of the input string"""
        decimal_number = self.str_to_int(str(request))
        alphabet, base36 = ['0123456789abcdefghijklmnopqrstuvwxyz', '']

        while decimal_number:
            decimal_number, i = divmod(decimal_number, 36)
            base36 = alphabet[i] + base36

        return base36 or alphabet[0]



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
This customization adds two new parameters to the ``ec2 run-instance``
command.  The first, ``--secondary-private-ip-addresses`` allows a list
of IP addresses within the specified subnet to be associated with the
new instance.  The second, ``--secondary-ip-address-count`` allows you
to specify how many additional IP addresses you want but the actual
address will be assigned for you.

This functionality (and much more) is also available using the
``--network-interfaces`` complex argument.  This just makes two of
the most commonly used features available more easily.
"""

from awscli.arguments import CustomArgument
from awscli.customizations.exceptions import ParamValidationError

# --secondary-private-ip-address
SECONDARY_PRIVATE_IP_ADDRESSES_DOCS = (
    '[EC2-VPC] A secondary private IP address for the network interface '
    'or instance. You can specify this multiple times to assign multiple '
    'secondary IP addresses.  If you want additional private IP addresses '
    'but do not need a specific address, use the '
    '--secondary-private-ip-address-count option.'
)

# --secondary-private-ip-address-count
SECONDARY_PRIVATE_IP_ADDRESS_COUNT_DOCS = (
    '[EC2-VPC] The number of secondary IP addresses to assign to '
    'the network interface or instance.'
)

# --associate-public-ip-address
ASSOCIATE_PUBLIC_IP_ADDRESS_DOCS = (
    '[EC2-VPC] If specified a public IP address will be assigned '
    'to the new instance in a VPC.'
)


def _add_params(argument_table, **kwargs):
    arg = SecondaryPrivateIpAddressesArgument(
        name='secondary-private-ip-addresses',
        help_text=SECONDARY_PRIVATE_IP_ADDRESSES_DOCS,
    )
    argument_table['secondary-private-ip-addresses'] = arg
    arg = SecondaryPrivateIpAddressCountArgument(
        name='secondary-private-ip-address-count',
        help_text=SECONDARY_PRIVATE_IP_ADDRESS_COUNT_DOCS,
    )
    argument_table['secondary-private-ip-address-count'] = arg
    arg = AssociatePublicIpAddressArgument(
        name='associate-public-ip-address',
        help_text=ASSOCIATE_PUBLIC_IP_ADDRESS_DOCS,
        action='store_true',
        group_name='associate_public_ip',
    )
    argument_table['associate-public-ip-address'] = arg
    arg = NoAssociatePublicIpAddressArgument(
        name='no-associate-public-ip-address',
        help_text=ASSOCIATE_PUBLIC_IP_ADDRESS_DOCS,
        action='store_false',
        group_name='associate_public_ip',
    )
    argument_table['no-associate-public-ip-address'] = arg


def _check_args(parsed_args, **kwargs):
    # This function checks the parsed args.  If the user specified
    # the --network-interfaces option with any of the scalar options we
    # raise an error.
    arg_dict = vars(parsed_args)
    if arg_dict['network_interfaces']:
        for key in (
            'secondary_private_ip_addresses',
            'secondary_private_ip_address_count',
            'associate_public_ip_address',
        ):
            if arg_dict[key]:
                msg = (
                    'Mixing the --network-interfaces option '
                    'with the simple, scalar options is '
                    'not supported.'
                )
                raise ParamValidationError(msg)


def _fix_args(params, **kwargs):
    # The RunInstances request provides some parameters
    # such as --subnet-id and --security-group-id that can be specified
    # as separate options only if the request DOES NOT include a
    # NetworkInterfaces structure.  In those cases, the values for
    # these parameters must be specified inside the NetworkInterfaces
    # structure.  This function checks for those parameters
    # and fixes them if necessary.
    # NOTE: If the user is a default VPC customer, RunInstances
    # allows them to specify the security group by name or by id.
    # However, in this scenario we can only support id because
    # we can't place a group name in the NetworkInterfaces structure.
    network_interface_params = [
        'PrivateIpAddresses',
        'SecondaryPrivateIpAddressCount',
        'AssociatePublicIpAddress',
    ]
    if 'NetworkInterfaces' in params:
        interface = params['NetworkInterfaces'][0]
        if any(param in interface for param in network_interface_params):
            if 'SubnetId' in params:
                interface['SubnetId'] = params['SubnetId']
                del params['SubnetId']
            if 'SecurityGroupIds' in params:
                interface['Groups'] = params['SecurityGroupIds']
                del params['SecurityGroupIds']
            if 'PrivateIpAddress' in params:
                ip_addr = {
                    'PrivateIpAddress': params['PrivateIpAddress'],
                    'Primary': True,
                }
                interface['PrivateIpAddresses'] = [ip_addr]
                del params['PrivateIpAddress']
            if 'Ipv6AddressCount' in params:
                interface['Ipv6AddressCount'] = params['Ipv6AddressCount']
                del params['Ipv6AddressCount']
            if 'Ipv6Addresses' in params:
                interface['Ipv6Addresses'] = params['Ipv6Addresses']
                del params['Ipv6Addresses']
            if 'EnablePrimaryIpv6' in params:
                interface['PrimaryIpv6'] = params['EnablePrimaryIpv6']
                del params['EnablePrimaryIpv6']


EVENTS = [
    ('building-argument-table.ec2.run-instances', _add_params),
    ('operation-args-parsed.ec2.run-instances', _check_args),
    ('before-parameter-build.ec2.RunInstances', _fix_args),
]


def register_runinstances(event_handler):
    # Register all of the events for customizing BundleInstance
    for event, handler in EVENTS:
        event_handler.register(event, handler)


def _build_network_interfaces(params, key, value):
    # Build up the NetworkInterfaces data structure
    if 'NetworkInterfaces' not in params:
        params['NetworkInterfaces'] = [{'DeviceIndex': 0}]

    if key == 'PrivateIpAddresses':
        if 'PrivateIpAddresses' not in params['NetworkInterfaces'][0]:
            params['NetworkInterfaces'][0]['PrivateIpAddresses'] = value
    else:
        params['NetworkInterfaces'][0][key] = value


class SecondaryPrivateIpAddressesArgument(CustomArgument):
    def add_to_parser(self, parser, cli_name=None):
        parser.add_argument(
            self.cli_name, dest=self.py_name, default=self._default, nargs='*'
        )

    def add_to_params(self, parameters, value):
        if value:
            value = [{'PrivateIpAddress': v, 'Primary': False} for v in value]
            _build_network_interfaces(parameters, 'PrivateIpAddresses', value)


class SecondaryPrivateIpAddressCountArgument(CustomArgument):
    def add_to_parser(self, parser, cli_name=None):
        parser.add_argument(
            self.cli_name, dest=self.py_name, default=self._default, type=int
        )

    def add_to_params(self, parameters, value):
        if value:
            _build_network_interfaces(
                parameters, 'SecondaryPrivateIpAddressCount', value
            )


class AssociatePublicIpAddressArgument(CustomArgument):
    def add_to_params(self, parameters, value):
        if value is True:
            _build_network_interfaces(
                parameters, 'AssociatePublicIpAddress', value
            )


class NoAssociatePublicIpAddressArgument(CustomArgument):
    def add_to_params(self, parameters, value):
        if value is False:
            _build_network_interfaces(
                parameters, 'AssociatePublicIpAddress', value
            )



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
import logging
import os

from awscli import argprocess
from awscli.compat import compat_open

logger = logging.getLogger(__name__)


class ResourceLoadingError(Exception):
    pass


def register_uri_param_handler(session, **kwargs):
    prefix_map = copy.deepcopy(LOCAL_PREFIX_MAP)
    handler = URIArgumentHandler(prefix_map)
    session.register('load-cli-arg', handler)


class URIArgumentHandler:
    def __init__(self, prefixes):
        self._prefixes = prefixes

    def __call__(self, event_name, param, value, **kwargs):
        """Handler that supports param values from local files."""
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        try:
            return get_paramfile(value, self._prefixes)
        except ResourceLoadingError as e:
            raise argprocess.ParamError(param.cli_name, str(e))


def get_paramfile(path, cases):
    """Load parameter based on a resource URI.

    It is possible to pass parameters to operations by referring
    to files or URI's.  If such a reference is detected, this
    function attempts to retrieve the data from the file or URI
    and returns it.  If there are any errors or if the ``path``
    does not appear to refer to a file or URI, a ``None`` is
    returned.

    :type path: str
    :param path: The resource URI, e.g. file://foo.txt.  This value
        may also be a non resource URI, in which case ``None`` is returned.

    :type cases: dict
    :param cases: A dictionary of URI prefixes to function mappings
        that a parameter is checked against.

    :return: The loaded value associated with the resource URI.
        If the provided ``path`` is not a resource URI, then a
        value of ``None`` is returned.

    """
    data = None
    if isinstance(path, str):
        for prefix, function_spec in cases.items():
            if path.startswith(prefix):
                function, kwargs = function_spec
                data = function(prefix, path, **kwargs)
    return data


def get_file(prefix, path, mode):
    file_path = os.path.expandvars(os.path.expanduser(path[len(prefix) :]))
    try:
        with compat_open(file_path, mode) as f:
            return f.read()
    except UnicodeDecodeError:
        raise ResourceLoadingError(
            'Unable to load paramfile (%s), text contents could '
            'not be decoded.  If this is a binary file, please use the '
            'fileb:// prefix instead of the file:// prefix.' % file_path
        )
    except OSError as e:
        raise ResourceLoadingError(
            'Unable to load paramfile %s: %s' % (path, e)
        )


LOCAL_PREFIX_MAP = {
    'file://': (get_file, {'mode': 'r'}),
    'fileb://': (get_file, {'mode': 'rb'}),
}



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from prompt_toolkit.application import get_app
from prompt_toolkit.filters import Condition, has_focus
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from awscli.customizations.wizard.exceptions import BaseWizardException
from awscli.customizations.wizard.ui.utils import (
    get_ui_control_by_buffer_name,
    move_to_previous_prompt,
    refresh_details_view,
    show_details_if_visible_by_default,
)


@Condition
def details_visible():
    return get_app().details_visible


@Condition
def save_details_visible():
    return get_app().save_details_visible


@Condition
def prompt_has_details():
    return get_app().traverser.current_prompt_has_details()


@Condition
def error_bar_enabled():
    return get_app().error_bar_visible is not None


def get_default_keybindings():
    kb = KeyBindings()

    @kb.add(Keys.ControlC)
    def exit_(event):
        event.app.exit(exception=KeyboardInterrupt())

    def submit_current_answer(event):
        current_prompt = event.app.traverser.get_current_prompt()
        prompt_buffer = get_ui_control_by_buffer_name(
            event.app.layout, current_prompt
        ).buffer
        try:
            event.app.traverser.submit_prompt_answer(prompt_buffer.text)
            if isinstance(
                event.app.layout.error_bar.current_error, BaseWizardException
            ):
                event.app.layout.error_bar.clear()
        except BaseWizardException as e:
            event.app.layout.error_bar.display_error(e)
            return False
        return True

    @kb.add(Keys.Tab)
    @kb.add(Keys.Enter)
    def next_prompt(event):
        if not submit_current_answer(event):
            return
        traverser = event.app.traverser
        layout = event.app.layout
        event.app.details_visible = False
        new_prompt = traverser.next_prompt()
        if new_prompt == traverser.DONE:
            layout.focus(layout.run_wizard_dialog)
        else:
            next_control = get_ui_control_by_buffer_name(layout, new_prompt)
            show_details_if_visible_by_default(event.app, new_prompt)
            refresh_details_view(event.app, new_prompt)
            layout.focus(next_control)

    @kb.add(Keys.BackTab)
    def previous_prompt(event):
        if not submit_current_answer(event):
            return
        event.app.details_visible = False
        move_to_previous_prompt(event.app)

    @kb.add(Keys.F2)
    def focus_on_details_panel(event):
        if event.app.details_visible:
            layout = event.app.layout
            if (
                layout.current_buffer
                and layout.current_buffer.name == 'details_buffer'
            ):
                current_prompt = event.app.traverser.get_current_prompt()
                current_control = get_ui_control_by_buffer_name(
                    layout, current_prompt
                )
                layout.focus(current_control)
            else:
                details_buffer = layout.get_buffer_by_name('details_buffer')
                layout.focus(details_buffer)

    @kb.add(Keys.F3, filter=prompt_has_details)
    def show_details(event):
        event.app.details_visible = not event.app.details_visible
        layout = event.app.layout
        current_prompt = event.app.traverser.get_current_prompt()
        current_control = get_ui_control_by_buffer_name(layout, current_prompt)
        if not event.app.details_visible:
            layout.focus(current_control)
        else:
            refresh_details_view(event.app, current_prompt)

    @kb.add(Keys.ControlS, filter=prompt_has_details)
    def show_save_details_dialogue(event):
        if not event.app.details_visible:
            event.app.details_visible = True
            current_prompt = event.app.traverser.get_current_prompt()
            refresh_details_view(event.app, current_prompt)
        event.app.save_details_visible = True
        save_dialogue = get_ui_control_by_buffer_name(
            event.app.layout, 'save_details_dialogue'
        )
        event.app.layout.focus(save_dialogue)

    @kb.add(Keys.F4, filter=error_bar_enabled)
    def show_details(event):
        event.app.error_bar_visible = not event.app.error_bar_visible

    return kb



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import argparse
import io
import logging
import re

import jmespath
from botocore.utils import ArgumentGenerator

from awscli.autocomplete.local.fetcher import CliDriverFetcher
from awscli.formatter import get_formatter

LOG = logging.getLogger(__name__)


class OutputGetter:
    def __init__(self, driver):
        self._cli_driver_fetcher = CliDriverFetcher(driver)
        self._output_formats = driver.arg_table['output'].choices
        self._session = driver.session
        self._current_expression = None

    def get_output(self, parsed):
        operation_model = self._cli_driver_fetcher.get_operation_model(
            parsed.lineage, parsed.current_command
        )
        if operation_model:
            output_shape = getattr(operation_model, 'output_shape', None)
            if self._shape_has_members(output_shape):
                operation = ''.join(
                    [
                        part.capitalize()
                        for part in parsed.current_command.split('-')
                    ]
                )
                output, error_message = self._get_output(parsed)
                if error_message is not None:
                    return error_message
                query, error_message = self._get_query(parsed)
                if error_message is not None:
                    return error_message
                return self._get_display(
                    operation, output_shape, output, query
                )
        return 'No output'

    def _shape_has_members(self, shape):
        return shape and (
            getattr(shape, 'members', False) or getattr(shape, 'member', False)
        )

    def _get_output(self, parsed):
        error_message = None
        if parsed.current_param == 'output':
            if parsed.current_fragment in self._output_formats:
                return parsed.current_fragment, error_message
        session_output = self._session.get_config_variable('output')
        output = parsed.global_params.get('output') or session_output
        if output not in self._output_formats:
            error_message = (
                "Bad value for --output: %s\n\nValid values are: %s"
                % (output, ', '.join(self._output_formats))
            )
        return output, error_message

    def _get_query(self, parsed):
        error_message = None
        if parsed.current_param == 'query':
            query = parsed.current_fragment
        else:
            query = parsed.global_params.get('query')
        if query:
            # Because output example has only 1 element in any list if
            # user enters query like Groups[2] jmespath will return "null"
            # so we change all the numbers in brackets to 0 to keep the output
            # meaningful
            query = re.sub(r'([\{\[])\d+?([\}\]])', r'\g<1>0\g<2>', query)
            # In case of incorrect expression we return an error message
            # but we want to do it only in the case expression is really broken
            # and not during user types it so if expression has open bracket
            # with input at the end or trailing dot we remove this part before
            # parsing, for example:
            #
            # for "Groups." we will parse only "Groups"
            # for "Groups.Names[34" we will parse only "Groups.Names"
            query = re.sub(r'([\{\[])[^\]^\}]*$', '', query).strip('.')
        try:
            self._current_expression = jmespath.compile(query)
        except jmespath.exceptions.ParseError as e:
            error_message = "Bad value for --query: %s\n\n%s" % (query, str(e))
        except jmespath.exceptions.EmptyExpressionError:
            self._current_expression = None
        except Exception:
            LOG.debug('Exception caught in OutputGetter: ', exc_info=True)
        return self._current_expression, error_message

    def _get_display(self, operation, output_shape, output, query):
        args = argparse.Namespace(query=query, color='off')
        argument_generator = ArgumentGenerator(use_member_names=True)
        response = argument_generator.generate_skeleton(output_shape)
        formatter = get_formatter(output, args)
        stream = io.StringIO()
        formatter(operation, response, stream)
        return stream.getvalue().replace('\t', '  ')



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.commands import BasicCommand
from awscli.customizations.dynamodb.subcommands import (
    PutCommand,
    SelectCommand,
)


def register_ddb(events):
    events.register("building-command-table.main", add_ddb)


def add_ddb(command_table, session, **kwargs):
    command_table['ddb'] = DDB(session)


class DDB(BasicCommand):
    NAME = 'ddb'
    DESCRIPTION = 'High level DynamoDB commands.'
    SYNOPSIS = "aws ddb <Command> [<Arg> ...]"
    SUBCOMMANDS = [
        {'name': 'select', 'command_class': SelectCommand},
        {'name': 'put', 'command_class': PutCommand},
    ]

    def _run_main(self, parsed_args, parsed_globals):
        if parsed_args.subcommand is None:
            self._raise_usage_error()



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.credentials import JSONFileCache
from botocore.exceptions import ProfileNotFound, UnknownCredentialError

from awscli.customizations.sso.login import LoginCommand
from awscli.customizations.sso.logout import LogoutCommand
from awscli.customizations.sso.utils import AWS_CREDS_CACHE_DIR


def register_sso_commands(event_emitter):
    event_emitter.register(
        'building-command-table.sso',
        add_sso_commands,
    )
    event_emitter.register(
        'session-initialized',
        inject_json_file_cache,
        unique_id='inject_sso_json_file_cache',
    )


def add_sso_commands(command_table, session, **kwargs):
    command_table['login'] = LoginCommand(session)
    command_table['logout'] = LogoutCommand(session)


def inject_json_file_cache(session, **kwargs):
    try:
        cred_chain = session.get_component('credential_provider')
        sso_provider = cred_chain.get_provider('sso')
        sso_provider.cache = JSONFileCache(AWS_CREDS_CACHE_DIR)
    except (ProfileNotFound, UnknownCredentialError):
        return



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Core planner and executor for wizards."""

import json
import os
import re
from functools import partial

import jmespath
from botocore import xform_name

from awscli.customizations.wizard.exceptions import (
    InvalidDataTypeConversionException,
)
from awscli.utils import json_encoder

DONE_SECTION_NAME = '__DONE__'
OUTPUT_SECTION_NAME = '__OUTPUT__'


class Runner:
    def __init__(self, planner, executor):
        self._planner = planner
        self._executor = executor

    def run(self, wizard_spec):
        params = self._planner.plan(wizard_spec['plan'])
        self._executor.execute(wizard_spec['execute'], params)


class Planner:
    _DONE_STEP = 'DONE'
    _STOP_RUNNING = object()

    def __init__(self, step_handlers):
        self._step_handlers = step_handlers
        # These will hold the values we store from
        # running through the plan steps.
        self._parameters = {}

    def plan(self, plan):
        self._parameters.clear()
        steps_iter = self._steps(plan)
        step = next(steps_iter)
        while step is not self._STOP_RUNNING:
            next_step_name = self._run_plan_step(step)
            step = steps_iter.send(next_step_name)
        return self._parameters

    def _steps(self, plan):
        step_names = list(plan)
        step_index = 0
        step_name = step_names[step_index]
        while True:
            next_step_name = yield plan[step_name]
            if next_step_name == self._DONE_STEP:
                yield self._STOP_RUNNING
            elif next_step_name is None:
                # If no explicit step is given, we go on to the next
                # sequential step.
                step_index += 1
                if step_index == len(step_names):
                    # If this is the last step, we're done with the
                    # plan phase.
                    yield self._STOP_RUNNING
                step_name = step_names[step_index]
            else:
                # If an explicit name is given we find the index
                # of the step name and update accordingly.
                step_index = step_names.index(next_step_name)
                step_name = next_step_name

    def _run_plan_step(self, step):
        # Running step consists of first fetching any values
        # based on what's defined in `values`.
        for key, value in step['values'].items():
            step_type = value['type']
            handler = self._step_handlers[step_type]
            response = handler.run_step(value, self._parameters)
            self._parameters[key] = response
        # And next we need to figure out if there's a next
        # step we need to resolve.  If the key is not defined
        # for this step we return None which results in moving
        # on to the next sequential step.
        if 'next_step' not in step:
            return
        return self._decide_next_step_name(step['next_step'])

    def _decide_next_step_name(self, next_step):
        if isinstance(next_step, str):
            # If an explicit step name is provided, we don't need to do
            # anything further.
            return next_step
        if 'switch' in next_step:
            varname = next_step['switch']
            value = self._parameters.get(varname)
            return next_step[value]


class BaseStep:
    # Subclasses must implement this.  This defines the name you'd
    # use for the `type` in a wizard definition.
    NAME = ''

    def run_step(self, step_definition, parameters):
        raise NotImplementedError("run_step")


class StaticStep(BaseStep):
    NAME = 'static'

    def run_step(self, step_definition, parameters):
        return step_definition['value']


class PromptStep(BaseStep):
    NAME = 'prompt'

    def __init__(self, prompter):
        self._prompter = prompter
        self._conversion_funcs = {
            'int': int,
            'float': float,
            'str': str,
            'bool': lambda x: True if x.lower() == 'true' else False,
        }

    def run_step(self, step_definition, parameters):
        choices = self._get_choices(step_definition, parameters)
        response = self._prompter.prompt(
            step_definition['description'], choices=choices
        )
        return self._convert_data_type_if_needed(response, step_definition)

    def _get_choices(self, step_definition, parameters):
        choices = step_definition.get('choices')
        if choices is not None:
            if isinstance(choices, str):
                # If 'choices' is a string, we assume it's the name of
                # a variable.
                return parameters[choices]
            return choices

    def _convert_data_type_if_needed(self, response, step_definition):
        if 'datatype' not in step_definition:
            return response
        return self._conversion_funcs[step_definition['datatype']](response)


class YesNoPrompt(PromptStep):
    """Shorthand for a yes/no prompt.

    Asking yes/no questions is common enough in wizards that
    this class provides a shorthand for doing this instead of having
    to write out the long form version using the general ``prompt``
    step.

    """

    NAME = 'yesno-prompt'

    def run_step(self, step_definition, parameters):
        choices = [
            {'display': 'Yes', 'actual_value': 'yes'},
            {'display': 'No', 'actual_value': 'no'},
        ]
        if step_definition.get('start_value', 'yes') == 'no':
            # They want the "No" choice to be the starting value so we
            # need to reverse the choices.
            choices[:] = choices[::-1]
        response = self._prompter.prompt(
            step_definition['question'], choices=choices
        )
        return response


class FilePromptStep(BaseStep):
    NAME = 'fileprompt'

    def __init__(self, prompter):
        self._prompter = prompter

    def run_step(self, step_definition, parameters):
        response = self._prompter.prompt(step_definition['description'])
        return os.path.expanduser(os.path.abspath(response))


class TemplateStep(BaseStep):
    NAME = 'template'

    CONDITION_PATTERN = re.compile(
        r'(?:^[ \t]*)?{%\s*if\s+(?P<condition>.+?)\s+%}(?:\s*[$|\n])?'
        r'(?P<body>.+?)[ \t]*{%\s*endif\s*%}[$|\n]?',
        re.DOTALL | re.MULTILINE | re.IGNORECASE,
    )
    _SUPPORTED_CONDITION_OPERATORS = [
        '==',
        '!=',
    ]

    def _check_condition(self, parameters, matchobj):
        group_dict = matchobj.groupdict()
        condition = group_dict['condition'].strip()
        for operator in self._SUPPORTED_CONDITION_OPERATORS:
            if operator in condition:
                condition = self._resolve_variables_in_condition(
                    condition, parameters
                )
                right, left = condition.split(operator, 1)
                right = right.strip()
                left = left.strip()
                if operator == '==':
                    if left == right:
                        return group_dict['body']
                elif operator == '!=':
                    if left != right:
                        return group_dict['body']
                return ''
        return group_dict['body']

    def _resolve_variables_in_condition(self, condition, parameters):
        try:
            return condition.format_map(parameters)
        except KeyError:
            return condition

    def _evaluate_conditions(self, value, parameters):
        condition_checker = partial(self._check_condition, parameters)
        return re.sub(self.CONDITION_PATTERN, condition_checker, value)

    def run_step(self, step_definition, parameters):
        value = self._evaluate_conditions(step_definition['value'], parameters)
        return value.format_map(parameters)


class APICallStep(BaseStep):
    NAME = 'apicall'

    def __init__(self, api_invoker):
        self._api_invoker = api_invoker

    def run_step(self, step_definition, parameters):
        service, op_name = step_definition['operation'].split('.', 1)
        return self._api_invoker.invoke(
            service=service,
            operation=op_name,
            api_params=step_definition['params'],
            plan_variables=parameters,
            optional_api_params=step_definition.get('optional_params'),
            query=step_definition.get('query'),
            cache=step_definition.get('cache', False),
            paginate=step_definition.get('paginate', False),
        )


class SharedConfigStep(BaseStep):
    NAME = 'sharedconfig'

    def __init__(self, config_api):
        self._config_api = config_api

    def run_step(self, step_definition, parameters):
        if step_definition['operation'] == 'ListProfiles':
            return self._config_api.list_profiles()
        elif step_definition['operation'] == 'GetValue':
            return self._config_api.get_value(
                profile=step_definition['params'].get('profile'),
                value=step_definition['params']['value'],
            )


class LoadDataStep(BaseStep):
    NAME = 'load-data'

    def run_step(self, step_definition, parameters):
        var_resolver = VariableResolver()
        value = var_resolver.resolve_variables(
            parameters,
            step_definition['value'],
        )
        load_type = step_definition['load_type']
        if load_type == 'json':
            return json.loads(value)
        else:
            raise ValueError(f'Unsupported load_type: {load_type}')


class DumpDataStep(BaseStep):
    NAME = 'dump-data'

    def run_step(self, step_definition, parameters):
        var_resolver = VariableResolver()
        value = var_resolver.resolve_variables(
            parameters,
            step_definition['value'],
        )
        dump_type = step_definition['dump_type']
        if dump_type == 'json':
            return json.dumps(value)
        else:
            raise ValueError(f'Unsupported load_type: {dump_type}')


class VariableResolver:
    _VAR_MATCH = re.compile(r'^{(.*?)}$')

    def resolve_variables(self, variables, params):
        """Replace references to variables with their values.

        Example::

            VariableResolver().resolve_variables(
              {'foo': 'bar'},
              {'MyValue': "{foo}"},
            ) == {'MyValue': 'bar'}

        """
        return self._resolve_vars(params, variables)

    def _resolve_vars(self, value, plan_vars):
        if isinstance(value, str):
            is_variable_ref = self._VAR_MATCH.search(value)
            if is_variable_ref:
                varname = is_variable_ref.group(1)
                return plan_vars[varname]
            return value
        elif isinstance(value, list):
            final = []
            for v in value:
                final.append(self._resolve_vars(v, plan_vars))
            return final
        elif isinstance(value, dict):
            final = {}
            for k, v in value.items():
                final[k] = self._resolve_vars(v, plan_vars)
            final = self._resolve_functions(final)
            return final
        else:
            return value

    def _resolve_functions(self, value):
        if len(value) != 1:
            return value
        only_key = list(value)[0]
        # The built in functions will likely move to another module
        # as we start to add more.
        if only_key == '__wizard__:File':
            filename = os.path.expanduser(value[only_key]['path'])
            with open(filename, 'rb') as f:
                return f.read()
        return value


class APIInvoker:
    """This class contains shared logic for the apicall step.

    The ``apicall`` in the planner and executor are slightly different
    but share a lot of similar logic.  This class is used share logic
    between the two steps.

    """

    def __init__(self, session):
        self._session = session
        self._response_cache = {}

    def invoke(
        self,
        service,
        operation,
        api_params,
        plan_variables,
        optional_api_params=None,
        query=None,
        cache=False,
        paginate=False,
    ):
        # TODO: All of the params that come from prompting the user
        # are strings.  We need a way to convert values to their
        # appropriate types.  We can either add typing into the wizard
        # spec or we possibly auto-convert based on the service
        # model (or both).
        resolved_params = self._resolve_params(
            api_params, optional_api_params, plan_variables
        )
        if cache:
            response = self._get_cached_api_call(
                service, operation, resolved_params, paginate
            )
        else:
            response = self._make_api_call(
                service, operation, resolved_params, paginate
            )
        if query is not None:
            response = jmespath.search(query, response)
        return response

    def _resolve_params(self, api_params, optional_params, plan_vars):
        resolver = VariableResolver()
        api_params_resolved = resolver.resolve_variables(plan_vars, api_params)
        if optional_params is not None:
            optional_params_resolved = resolver.resolve_variables(
                plan_vars, optional_params
            )
            for key, value in optional_params_resolved.items():
                if key not in api_params_resolved and value is not None:
                    api_params_resolved[key] = value
        return api_params_resolved

    def _make_api_call(self, service, operation, resolved_params, paginate):
        client = self._session.create_client(service)
        client_method_name = xform_name(operation)
        if paginate:
            paginator = client.get_paginator(client_method_name)
            return paginator.paginate(**resolved_params).build_full_result()
        else:
            return getattr(client, client_method_name)(**resolved_params)

    def _get_cached_api_call(
        self, service, operation, resolved_params, paginate
    ):
        cache_key = self._get_cache_key(service, operation, resolved_params)
        if cache_key not in self._response_cache:
            response = self._make_api_call(
                service, operation, resolved_params, paginate
            )
            self._response_cache[cache_key] = response
        return self._response_cache[cache_key]

    def _get_cache_key(self, service_name, operation, resolved_params):
        return (
            service_name,
            operation,
            json.dumps(resolved_params, default=json_encoder),
        )


class Executor:
    def __init__(self, step_handlers):
        self._step_handlers = step_handlers

    def execute(self, step, parameters):
        # We may eventually support jumping around to different step
        # names, but for now, we just iterate through the steps in order.
        for group in step.values():
            for step in group:
                self._execute_step(step, parameters)

    def _execute_step(self, step, parameters):
        if 'condition' in step:
            should_run = ConditionEvaluator().evaluate(
                step['condition'], parameters
            )
            if not should_run:
                return
        step_type = step['type']
        handler = self._step_handlers[step_type]
        handler.run_step(step, parameters)


class ConditionEvaluator:
    def evaluate(self, condition, parameters):
        statuses = []
        if not isinstance(condition, list):
            condition = [condition]
        for single in condition:
            statuses.append(self._check_single_condition(single, parameters))
        return all(statuses)

    def _check_single_condition(self, single, parameters):
        varname = single['variable']
        if 'equals' in single:
            expected = single['equals']
            return parameters.get(varname) == expected
        return False


class DataTypeConverter:
    CONVERSION_FUNCS = {
        'int': int,
        'float': float,
        'str': str,
        'bool': lambda x: x.lower() == 'true',
    }

    @classmethod
    def convert(cls, datatype, value):
        try:
            return cls.CONVERSION_FUNCS[datatype](value)
        except ValueError:
            raise InvalidDataTypeConversionException(value, datatype)


class ExecutorStep:
    # Subclasses must implement this to specify what name to use
    # for the `type` in a wizard definition.
    NAME = ''

    def run_step(self, step_definition, parameters):
        raise NotImplementedError("run_step")


class APICallExecutorStep(ExecutorStep):
    NAME = 'apicall'

    def __init__(self, api_invoker):
        self._api_invoker = api_invoker

    def run_step(self, step_definition, parameters):
        service, op_name = step_definition['operation'].split('.', 1)
        response = self._api_invoker.invoke(
            service=service,
            operation=op_name,
            api_params=step_definition['params'],
            plan_variables=parameters,
            optional_api_params=step_definition.get('optional_params'),
            query=step_definition.get('query'),
        )
        if 'output_var' in step_definition:
            parameters[step_definition['output_var']] = response


class SharedConfigExecutorStep(ExecutorStep):
    NAME = 'sharedconfig'

    def __init__(self, config_api):
        self._config_api = config_api

    def run_step(self, step_definition, parameters):
        config_params = {}
        profile = None
        if 'profile' in step_definition:
            profile = self._resolve_params(
                step_definition['profile'], parameters
            )
        config_params = self._resolve_params(
            step_definition['params'], parameters
        )
        self._config_api.set_values(config_params, profile=profile)

    def _resolve_params(self, value, params):
        return VariableResolver().resolve_variables(params, value)


class SharedConfigAPI:
    """Simplified interface to reading/writing the ~/.aws/config file.

    This provides a simplified interface over the config file writer
    and the config operations provided by a botocore session.

    This allows similar logic to be shared by the planner and executor.

    """

    def __init__(self, session, config_writer):
        self._session = session
        self._config_writer = config_writer

    def list_profiles(self):
        return self._session.available_profiles

    def get_value(self, value, profile=None):
        # TODO: handle profile
        return self._session.get_config_variable(value)

    def set_values(self, values, profile=None):
        config_params = {}
        if profile is not None:
            section = profile
            if profile != 'default':
                section = 'profile %s' % section
            config_params['__section__'] = section
        config_params.update(values)
        config_filename = os.path.expanduser(
            self._session.get_config_variable('config_file')
        )
        self._config_writer.update_config(config_params, config_filename)


class DefineVariableStep(ExecutorStep):
    NAME = 'define-variable'

    def run_step(self, step_definition, parameters):
        value = step_definition['value']
        resolved_value = VariableResolver().resolve_variables(
            parameters, value
        )
        key = step_definition['varname']
        parameters[key] = resolved_value


class MergeDictStep(ExecutorStep):
    NAME = 'merge-dict'

    def run_step(self, step_definition, parameters):
        var_resolver = VariableResolver()
        result = {}
        for overlay in step_definition['overlays']:
            resolved_overlay = var_resolver.resolve_variables(
                parameters,
                overlay,
            )
            result = self._deep_merge(result, resolved_overlay)
        parameters[step_definition['output_var']] = result

    def _deep_merge(self, original, newvalue):
        if isinstance(newvalue, list) and isinstance(original, list):
            # We want to concat list together during a deep merge.
            return original + newvalue
        elif isinstance(newvalue, dict) and isinstance(original, dict):
            result = original.copy()
            for key, value in newvalue.items():
                if key not in result:
                    result[key] = value
                else:
                    result[key] = self._deep_merge(original[key], value)
            return result
        else:
            return newvalue


class LoadDataExecutorStep(ExecutorStep):
    NAME = 'load-data'

    def run_step(self, step_definition, parameters):
        loaded_value = LoadDataStep().run_step(step_definition, parameters)
        parameters[step_definition['output_var']] = loaded_value


class DumpDataExecutorStep(ExecutorStep):
    NAME = 'dump-data'

    def run_step(self, step_definition, parameters):
        dumped_value = DumpDataStep().run_step(step_definition, parameters)
        parameters[step_definition['output_var']] = dumped_value



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import errno
import json
import logging
import subprocess

import awscli
from awscli.compat import ignore_user_entered_signals
from awscli.customizations.commands import BasicCommand
from awscli.customizations.lightsail import helptext

logger = logging.getLogger(__name__)

ERROR_MESSAGE = (
    'The Lightsail Control (lightsailctl) plugin was not found. ',
    'To download and install it, see ',
    'https://lightsail.aws.amazon.com/ls/docs/en_us/articles/amazon-lightsail-install-software',
)

INPUT_VERSION = '1'


class PushContainerImage(BasicCommand):
    NAME = 'push-container-image'

    DESCRIPTION = 'Push container image for use in a service deployment.'

    ARG_TABLE = [
        {
            'name': 'service-name',
            'help_text': helptext.SERVICENAME,
            'required': True,
        },
        {'name': 'image', 'help_text': helptext.IMAGE, 'required': True},
        {'name': 'label', 'help_text': helptext.LABEL, 'required': True},
    ]

    def _run_main(self, parsed_args, parsed_globals):
        payload = self._get_input_request(parsed_args, parsed_globals)
        try:
            # ignore_user_entered_signals ignores these signals
            # because if signals which kills the process are not
            # captured would kill the foreground process but not the
            # background one. Capturing these would prevents process
            # from getting killed and these signals are input to plugin
            # and handling in there
            with ignore_user_entered_signals():
                # call executable with necessary input
                subprocess.run(
                    ['lightsailctl', '--plugin', '--input-stdin'],
                    input=json.dumps(payload).encode('utf-8'),
                    check=True,
                )
            return 0
        except OSError as ex:
            if ex.errno == errno.ENOENT:
                logger.debug('lightsailctl not found', exc_info=True)
                raise ValueError(''.join(ERROR_MESSAGE))

    def _get_input_request(self, parsed_args, parsed_globals):
        input_request = {
            'inputVersion': INPUT_VERSION,
            'operation': 'PushContainerImage',
        }

        payload = dict(
            service=parsed_args.service_name,
            image=parsed_args.image,
            label=parsed_args.label,
        )

        configuration = {}
        configuration['debug'] = parsed_globals.debug

        if parsed_globals.endpoint_url is not None:
            configuration['endpoint'] = parsed_globals.endpoint_url

        if parsed_globals.verify_ssl is not None:
            configuration['doNotVerifySSL'] = not parsed_globals.verify_ssl

        configuration['paginate'] = parsed_globals.paginate

        if parsed_globals.output is not None:
            configuration['output'] = parsed_globals.output

        if parsed_globals.query is not None:
            configuration['query'] = parsed_globals.query

        if parsed_globals.profile:
            configuration['profile'] = parsed_globals.profile
        elif self._session.get_config_variable('profile'):
            configuration['profile'] = self._session.get_config_variable(
                'profile'
            )

        if parsed_globals.region:
            configuration['region'] = parsed_globals.region
        elif self._session.get_config_variable('region'):
            configuration['region'] = self._session.get_config_variable(
                'region'
            )

        configuration['doNotSignRequest'] = not parsed_globals.sign_request

        if parsed_globals.ca_bundle:
            configuration['caBundle'] = parsed_globals.ca_bundle
        elif self._session.get_config_variable('ca_bundle'):
            configuration['caBundle'] = self._session.get_config_variable(
                'ca_bundle'
            )

        if parsed_globals.read_timeout is not None:
            configuration['readTimeout'] = parsed_globals.read_timeout

        if parsed_globals.connect_timeout is not None:
            configuration['connectTimeout'] = parsed_globals.connect_timeout

        if parsed_globals.cli_binary_format is not None:
            configuration['cliBinaryFormat'] = parsed_globals.cli_binary_format

        configuration['cliVersion'] = awscli.__version__

        input_request['payload'] = payload
        input_request['configuration'] = configuration
        return input_request



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from ruamel.yaml import YAML

from awscli.customizations.commands import BasicCommand
from awscli.customizations.wizard.factory import create_wizard_app


def register_dev_commands(event_handlers):
    event_handlers.register(
        'building-command-table.cli-dev', WizardDev.add_command
    )


def create_default_wizard_dev_runner(session):
    return WizardDevRunner(
        wizard_loader=WizardLoader(),
        session=session,
    )


class WizardLoader:
    def load(self, contents):
        yaml = YAML(typ="rt")
        data = yaml.load(contents)
        return data


class WizardDevRunner:
    def __init__(self, wizard_loader, session):
        self._wizard_loader = wizard_loader
        self._session = session

    def run_wizard(self, wizard_contents):
        """Run a single wizard given the contents as a string."""
        loaded = self._wizard_loader.load(wizard_contents)
        app = create_wizard_app(loaded, self._session)
        app.run()
        print(f'Collected values: {app.values}')


class WizardDev(BasicCommand):
    NAME = 'wizard-dev'
    DESCRIPTION = (
        'Internal command from developing, testing and debugging wizards.\n'
        'This command is not intended for normal end usage. '
        'Do not rely on this command, backwards compatibility '
        'is not guaranteed.  This command may be removed in '
        'future versions.\n'
    )
    ARG_TABLE = [
        {
            'name': 'run-wizard',
            'help_text': 'Run a wizard given a wizard file.',
        }
    ]

    def __init__(self, session, dev_runner=None):
        super(WizardDev, self).__init__(session)
        if dev_runner is None:
            dev_runner = create_default_wizard_dev_runner(session)
        self._dev_runner = dev_runner

    def _run_main(self, args, parsed_globals):
        if args.run_wizard is not None:
            return self._run_wizard(args.run_wizard)

    def _run_wizard(self, wizard_contents):
        self._dev_runner.run_wizard(wizard_contents)
        return 0



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import shlex
import sys
from contextlib import contextmanager, nullcontext

from prompt_toolkit.application import Application
from prompt_toolkit.completion import Completer, Completion, ThreadedCompleter
from prompt_toolkit.document import Document

from awscli.autocomplete import parser
from awscli.autocomplete.local import model
from awscli.autoprompt.doc import DocsGetter
from awscli.autoprompt.factory import PromptToolkitFactory
from awscli.autoprompt.logger import PromptToolkitHandler
from awscli.autoprompt.output import OutputGetter
from awscli.logger import LOG_FORMAT, disable_crt_logging

LOG = logging.getLogger(__name__)


@contextmanager
def loggers_handler_switcher():
    old_handlers = {}
    loggers = [name for name in logging.root.manager.loggerDict]

    handler = PromptToolkitHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(LOG_FORMAT))
    for logger_name in loggers:
        logger = logging.getLogger(logger_name)
        old_handlers[logger_name] = logger.handlers
        logger.handlers = []
        logger.addHandler(handler)
    yield None
    for logger_name in loggers:
        logger = logging.getLogger(logger_name)
        logger.handlers = []
        for handler in old_handlers[logger_name]:
            logger.addHandler(handler)


class PromptToolkitPrompter:
    """Handles the actual prompting in the autoprompt workflow."""

    def __init__(
        self,
        completion_source,
        driver,
        completer=None,
        factory=None,
        app=None,
        cli_parser=None,
        output=None,
        app_input=None,
    ):
        self._completion_source = completion_source
        self._output = output
        self._input = app_input
        if completer is None:
            completer = PromptToolkitCompleter(self._completion_source)
        # We wrap our completer with a ThreadedCompleter to make autocompletion
        # generation more responsive.
        self._completer = ThreadedCompleter(completer)
        if factory is None:
            factory = PromptToolkitFactory(completer=self._completer)
        self._parser = cli_parser
        if self._parser is None:
            self._parser = parser.CLIParser(
                model.ModelIndex(), return_first_command_match=True
            )
        self._factory = factory
        self.input_buffer = None
        self.doc_buffer = None
        self.output_buffer = None
        if app is None:
            app = self.create_application()
        self.app = app
        self._args = []
        self._driver = driver
        self._docs_getter = DocsGetter(self._driver)
        self._output_getter = OutputGetter(self._driver)

    def args(self, value):
        self._args = value

    args = property(None, args)

    def _create_buffers(self):
        self.input_buffer = self._factory.create_input_buffer(
            self.update_bottom_buffers_text
        )
        self.doc_buffer = self._factory.create_doc_buffer()
        self.output_buffer = self._factory.create_output_buffer()

    def _create_containers(self):
        input_buffer_container = self._factory.create_input_buffer_container(
            self.input_buffer
        )
        doc_window = self._factory.create_searchable_window(
            'Doc panel', self.doc_buffer
        )
        output_window = self._factory.create_searchable_window(
            'Output panel', self.output_buffer
        )
        return input_buffer_container, doc_window, output_window

    def create_application(self):
        self._create_buffers()
        input_buffer_container, doc_window, output_window = (
            self._create_containers()
        )
        layout = self._factory.create_layout(
            on_input_buffer_text_changed=self.update_bottom_buffers_text,
            input_buffer_container=input_buffer_container,
            doc_window=doc_window,
            output_window=output_window,
        )
        kb_manager = self._factory.create_key_bindings()
        kb = kb_manager.keybindings
        app = Application(
            layout=layout,
            key_bindings=kb,
            full_screen=False,
            output=self._output,
            erase_when_done=True,
            input=self._input,
        )
        self._set_app_defaults(app)
        return app

    def _set_app_defaults(self, app):
        app.show_doc = False
        app.multi_column = False
        app.show_help = False
        app.debug = False
        app.show_output = False
        return app

    def update_bottom_buffers_text(self, *args):
        parsed = self._parser.parse('aws ' + self.input_buffer.document.text)
        self._update_doc_window_contents(parsed)
        self._update_output_window_contents(parsed)

    def _update_output_window_contents(self, parsed):
        content = self._output_getter.get_output(parsed)
        if content is not None and content != self.output_buffer.document.text:
            self.output_buffer.reset()
            new_document = Document(text=content, cursor_position=0)
            self.output_buffer.set_document(new_document, bypass_readonly=True)

    def _update_doc_window_contents(self, parsed):
        content = self._docs_getter.get_docs(parsed)
        # The only way to "modify" a read-only buffer in prompt_toolkit is to
        # create a new `prompt_toolkit.document.Document`. A 'cursor_position'
        # of 0 allows us to start viewing the documentation from the beginning,
        # as opposed to the default of moving the cursor to the end.
        #
        # Note: `content` may be None if we are still in the middle of typing a
        # command (e.g. `aws ec2 descr`) or if the current command is invalid.
        # In these cases, we don't update the doc window.
        if content is not None and content != self.doc_buffer.document.text:
            self.doc_buffer.reset()
            new_document = Document(text=content, cursor_position=0)
            self.doc_buffer.set_document(new_document, bypass_readonly=True)

    def _set_debug_mode(self):
        if '--debug' in self._args:
            self.app.debug = True

    def pre_run(self):
        if '--cli-auto-prompt' in self._args:
            self._args.remove('--cli-auto-prompt')
        # Before prompting the user for a command, we copy what they entered
        # at the command line into the buffer (aka "construction zone"), set
        # the cursor position, and make sure to immediately provide a pop-up
        # menu with autocompletion options.
        #
        # If what the user entered at the command line already has no
        # completions, we just keep it as it is without showing
        # any suggestions.
        self.input_buffer.insert_text(' '.join(self._args))
        cmd_line_text = 'aws ' + self.input_buffer.document.text
        self._set_input_buffer_text(cmd_line_text)
        self.update_bottom_buffers_text()
        self.input_buffer.start_completion()

    def _set_input_buffer_text(self, cmd_line_text):
        """If entered command line does not have trailing space and can not
        be autocompleted we assume that it is a completed part of command
        and add trailing space to it"""
        if cmd_line_text[-1] == ' ':
            return
        if self._can_autocomplete(cmd_line_text):
            return
        if cmd_line_text.strip() != 'aws':
            self.input_buffer.insert_text(' ')

    def _can_autocomplete(self, cmd_line_text):
        return bool(self._completion_source.autocomplete(cmd_line_text))

    def _quote_args_with_spaces(self, args):
        return [shlex.quote(arg) for arg in args]

    def prompt_for_args(self, original_args):
        """Prompt for values for a given CLI command.

        :type original_args: list
        :param original_args: A list of the entered command at the program entry
            point, less the initial `aws` command.

        :rtype: list
        :return: A list of what was entered into the autoprompt buffer,
            delimited by whitespace characters.

        """
        self._args = self._quote_args_with_spaces(original_args)
        self._set_debug_mode()
        logging_manager = nullcontext
        if self.app.debug:
            # NOTE: The CRT library is not integrated with the Python stdlib
            # logger and only allows you to write to file names and
            # stdout/stderr. Therefore we currently are unable to switch it
            # to write logs to our promptoolkit debug buffer. So instead of
            # letting it write to stderr and disrupting the UI, we are
            # disabling it while in auto-prompt mode until it has
            # integrations with the Python stdlib logger.
            disable_crt_logging()
            logging_manager = loggers_handler_switcher
        with logging_manager():
            self.app.run(pre_run=self.pre_run)
        cmd_line_text = self.input_buffer.document.text
        # Once the application is finished running, the screen is cleared.
        # Here, we display the command to be run so that the user knows what
        # command caused the output that they're seeing.
        sys.stdout.write(f'> aws {cmd_line_text}\n')
        return shlex.split(cmd_line_text)


class PromptToolkitCompleter(Completer):
    """Handles conversion of autocompletion data from our completion source
    into `prompt_toolkit.Completion` objects that prompt_toolkit understands.

    Note: prompt_toolkit expects a method `get_completions()` to generate the
    `prompt_toolkit.Completion` objects.

    """

    def __init__(self, completion_source):
        self._completion_source = completion_source

    def _convert_to_prompt_completions(
        self, low_level_completions, text_before_cursor
    ):
        # Converts the low-level completions from the model autocompleter
        # and converts them to Completion() objects that are used by
        # prompt_toolkit.
        word_before_cursor = self._strip_whitespace(text_before_cursor)
        completions = self._filter_completions(low_level_completions)
        ordered_completions = self._prioritize_required_args(completions)
        for completion in ordered_completions:
            display_text = self._get_display_text(completion)
            display_meta = self._get_display_meta(completion)
            location = self._get_starting_location_of_last_word(
                text_before_cursor, word_before_cursor
            )
            yield Completion(
                completion.name,
                location,
                display=display_text,
                display_meta=display_meta,
            )

    def get_completions(self, document, complete_event):
        try:
            text_before_cursor = document.text_before_cursor
            text_to_autocomplete = 'aws ' + text_before_cursor
            completions = self._completion_source.autocomplete(
                text_to_autocomplete, len(text_to_autocomplete)
            )
            yield from self._convert_to_prompt_completions(
                completions, text_before_cursor
            )
        except Exception as e:
            LOG.debug(
                'Exception caught in PromptToolkitCompleter: %s' % e,
                exc_info=True,
            )

    def _strip_whitespace(self, text):
        word_before_cursor = ''
        if text.strip():
            word_before_cursor = text.strip().split()[-1]
        return word_before_cursor

    def _prioritize_required_args(self, completions):
        required_args = self._get_required_args(completions)
        optional_args = self._get_optional_args(completions)
        return required_args + optional_args

    def _get_required_args(self, completions):
        results = [arg for arg in completions if arg.required]
        return results

    def _get_optional_args(self, completions):
        results = [arg for arg in completions if not arg.required]
        return results

    def _get_display_text(self, completion):
        if completion.display_text is not None:
            return completion.display_text
        display_text = completion.name
        if completion.name.startswith('--') and completion.required:
            display_text += ' (required)'
        return display_text

    def _get_display_meta(self, completion):
        display_meta = ''
        type_name = getattr(completion, 'cli_type_name', None)
        help_text = getattr(completion, 'help_text', None)
        if type_name:
            display_meta += f'[{type_name}] '
        if help_text:
            display_meta += f'{help_text}'
        return display_meta

    def _filter_completions(self, completions):
        completions = self._filter_out_autoprompt_overrides(completions)
        completions = self._remove_duplicate_completions(completions)
        return completions

    def _filter_out_autoprompt_overrides(self, completions):
        filtered_completions = [
            completion
            for completion in completions
            if completion.name
            not in ['--cli-auto-prompt', '--no-cli-auto-prompt']
        ]
        return filtered_completions

    def _remove_duplicate_completions(self, completions):
        unique_completion_names = []
        unique_completions = []
        for completion in completions:
            if completion.name not in unique_completion_names:
                unique_completion_names.append(completion.name)
                unique_completions.append(completion)
        return unique_completions

    def _get_starting_location_of_last_word(
        self, text_before_cursor, word_before_cursor
    ):
        if text_before_cursor and text_before_cursor[-1] == ' ':
            location = 0
        else:
            location = -len(word_before_cursor)
        return location
